from mot.model_building.cl_functions.parameters import FreeParameter
from mot.model_building.cl_functions.base import ModelFunction
from mot.model_building.cl_functions.library_functions import Bessel
from mot.model_building.parameter_functions.transformations import ClampTransform
from mot.cl_data_type import CLDataType


__author__ = 'Robbert Harms'
__date__ = "2014-08-05"
__license__ = "LGPL v3"
__maintainer__ = "Robbert Harms"
__email__ = "robbert.harms@maastrichtuniversity.nl"


class EvaluationModel(ModelFunction):

    def __init__(self, name, cl_function_name, parameter_list, dependency_list=()):
        """The evaluation model is the model under which you evaluate the estimated results against the data.

        This normally embed the noise model assumptions of your data.
        """
        super(EvaluationModel, self).__init__(name, cl_function_name, parameter_list, dependency_list=dependency_list)

    def get_objective_function(self, fname, inst_per_problem, eval_fname, obs_fname, param_listing):
        """Get the cl code for the objective function under the given noise model.

        Args:
            fname (str): the name of the resulting function
            inst_per_problem (int): the number of instances per problem
            eval_fname (str): the name of the function that can be called to get the evaluation, its signature is:

                .. code-block:: c

                    mot_float_type <fname>(const void* data, const mot_float_type* x,
                                           const int observation_index);

            obs_fname (str): the name of the function that can be called for the observed data, its signature is:

                .. code-block:: c

                    mot_float_type <fname>(const void* data, const int observation_index);

            param_listing (str): the parameter listings for the parameters of the noise model

        Returns:
            str: The objective function under this noise model, its signature is:

                .. code-block:: c

                    double <fname>(const void* const data, mot_float_type* const x);

            That is, it always returns a double since the summations may get large.
        """

    def get_objective_list_function(self, fname, inst_per_problem, eval_fname, obs_fname, param_listing):
        """Get the cl code for the objective function list under the given noise model.

        This function is used by some evaluation routines (like for example LevenbergMarquardt) that need
        a list of objective values (one per instance point), instead of a single objective function scalar.

        The returned CL function should accept a result vector in which the results per instance point ought
        to be placed.

        Args:
            fname (str): the name of the resulting function
            inst_per_problem (int): the number of instances per problem
            eval_fname (str): the name of the function that can be called to get the evaluation, its signature is:

                .. code-block:: c

                    mot_float_type <fname>(const void* data, const mot_float_type* x,
                                           const int observation_index);

            obs_fname (str): the name of the function that can be called for the observed data, its signature is:

                .. code-block:: c

                    mot_float_type <fname>(const void* data, const int observation_index);

            param_listing (str): the parameter listings for the parameters of the noise model

        Returns:
            str: The objective function under this noise model, its signature is:

                .. code-block:: c

                    void (const void* const data, mot_float_type* const x, mot_float_type* result);
        """

    def get_log_likelihood_function(self, fname, inst_per_problem, eval_fname, obs_fname, param_listing,
                                    full_likelihood=True):
        """Get the cl code for the log likelihood function under the given noise model.

        Args:
            fname (str): the name of the resulting function
            inst_per_problem (int): the number of instances per problem
            eval_fname (str): the name of the function that can be called to get the evaluation, its signature is:

                .. code-block:: c

                    mot_float_type <fname>(const void* data, const mot_float_type* x,
                                           const int observation_index);

            obs_fname (str): the name of the function that can be called for the observed data, its signature is:

                .. code-block:: c

                    mot_float_type <fname>(const void* data, const int observation_index);

            param_listing (str): the parameter listings for the parameters of the noise model
            full_likelihood (boolean): if we want the complete likelihood, or if we can drop the constant terms.
                The default is the complete likelihood. Disable for speed.

        Returns:
            str: the objective function under this noise model, its signature is:

                .. code-block:: c

                    double <fname>(const void* const data, mot_float_type* const x);

            That is, it always returns a double since the summations may get large.
        """

    def get_noise_std_param_name(self):
        """Get the name of the parameter that is associated with the noise standard deviation in the problem data.

        Returns:
            str: the name of the parameter that is associated with the noise_std in the problem data.
        """
        return 'sigma'


class SumOfSquares(EvaluationModel):

    def __init__(self):
        """Evaluates the distance between the estimated signal and the data using the sum of squared differences.

        This is implemented as::

            sum((observation - evaluation)^2)
        """
        super(EvaluationModel, self).__init__('SumOfSquaresNoise', 'sumOfSquaresNoise', (), ())

    def get_objective_function(self, fname, inst_per_problem, eval_fname, obs_fname, param_listing):
        return '''
            double ''' + fname + '''(const void* const data, mot_float_type* const x){
                ''' + param_listing + '''
                double sum = 0.0;
                for(int i = 0; i < ''' + str(inst_per_problem) + '''; i++){
                    sum += pown(''' + obs_fname + '''(data, i) - ''' + eval_fname + '''(data, x, i), 2);
                }
                return sum;
            }
        '''

    def get_log_likelihood_function(self, fname, inst_per_problem, eval_fname, obs_fname, param_listing,
                                    full_likelihood=True):
        return '''
            double ''' + fname + '''(const void* const data, const mot_float_type* const x){
                ''' + param_listing + '''
                double sum = 0.0;
                for(int i = 0; i < ''' + str(inst_per_problem) + '''; i++){
                    sum += pown(''' + obs_fname + '''(data, i) - ''' + eval_fname + '''(data, x, i), 2);
                }
                return - sum;
            }
        '''

    def get_objective_list_function(self, fname, inst_per_problem, eval_fname, obs_fname, param_listing):
        return '''
            void ''' + fname + '''(const void* const data, mot_float_type* const x, mot_float_type* result){
                ''' + param_listing + '''
                for(int i = 0; i < ''' + str(inst_per_problem) + '''; i++){
                    result[i] = ''' + obs_fname + '''(data, i) - ''' + eval_fname + '''(data, x, i);
                }
            }
        '''


class GaussianEvaluationModel(EvaluationModel):

    def __init__(self):
        """This uses the log of the Gaussian PDF for the maximum likelihood estimator and for the log likelihood.

        The PDF is defined as:

        .. code-block:: c

            PDF = 1/(sigma * sqrt(2*pi)) * exp(-(observation - evaluation)^2 / (2 * sigma^2))

        To have the joined probability over all instances one would normally have to take the product
        over all ``n`` instances:

        .. code-block:: c

            product(PDF)

        Instead of taking the product of this PDF we take the sum of the log of the PDF:

        .. code-block:: c

            sum(log(PDF))

        Where the log of the PDF is given by:

        .. code-block:: c

            log(PDF) = - ((observation - evaluation)^2 / (2 * sigma^2)) - log(sigma * sqrt(2*pi))


        For the maximum likelihood estimator we then need to use the negative of this sum:

        .. code-block:: c

            - sum(log(PDF)).
        """
        super(GaussianEvaluationModel, self).__init__(
            'GaussianNoise',
            'gaussianNoiseModel',
            (FreeParameter(CLDataType.from_string('mot_float_type'), 'sigma', True, 1, 0, 'INFINITY',
                           parameter_transform=ClampTransform()),), ())

    def get_objective_function(self, fname, inst_per_problem, eval_fname, obs_fname, param_listing):
        """Get the Gaussian objective function.

        This omits the constant terms for speed reasons. Omitted terms are:

         .. code-block:: c

            + log(GaussianNoise_sigma * sqrt(2 * M_PI))

        """
        return '''
            double ''' + fname + '''(const void* const data, mot_float_type* const x){
                ''' + param_listing + '''
                double sum = 0.0;
                for(int i = 0; i < ''' + str(inst_per_problem) + '''; i++){
                    sum += pown(''' + obs_fname + '''(data, i) - ''' + eval_fname + '''(data, x, i), 2);
                }
                return sum / (2 * GaussianNoise_sigma * GaussianNoise_sigma);
            }
        '''

    def get_objective_list_function(self, fname, inst_per_problem, eval_fname, obs_fname, param_listing):
        return '''
            void ''' + fname + '''(const void* const data, mot_float_type* const x, mot_float_type* result){
                ''' + param_listing + '''
                for(int i = 0; i < ''' + str(inst_per_problem) + '''; i++){
                    result[i] = ''' + obs_fname + '''(data, i) - ''' + eval_fname + '''(data, x, i);
                }
            }
        '''

    def get_log_likelihood_function(self, fname, inst_per_problem, eval_fname, obs_fname, param_listing,
                                    full_likelihood=True):
        return '''
            double ''' + fname + '''(const void* const data, const mot_float_type* const x){
                ''' + param_listing + '''
                double sum = 0.0;
                for(int i = 0; i < ''' + str(inst_per_problem) + '''; i++){
                    sum += pown(''' + obs_fname + '''(data, i) - ''' + eval_fname + '''(data, x, i), 2);
                }
                return - sum / (2 * GaussianNoise_sigma * GaussianNoise_sigma)
                    ''' + ('+' + str(inst_per_problem) + ' * log(GaussianNoise_sigma * sqrt(2 * M_PI))'
                           if full_likelihood else '') + ''';

            }
        '''


class OffsetGaussianEvaluationModel(EvaluationModel):

    def __init__(self):
        """This uses the log of the Gaussian PDF for the maximum likelihood estimator and for the log likelihood.

        The PDF is defined as:

        .. code-block:: c

            PDF = 1/(sigma * sqrt(2*pi)) * exp(-(observation - sqrt(evaluation^2 + sigma^2))^2 / (2 * sigma^2))

        To have the joined probability over all instances one would have to take the product over all n instances:

        .. code-block:: c

            product(PDF)

        Instead of taking the product of this PDF we take the sum of the log of the PDF:

        .. code-block:: c

            sum(log(PDF))

        Where the log of the PDF is given by:

        .. code-block:: c

            log(PDF) = - ((observation - sqrt(evaluation^2 + sigma^2))^2 / (2 * sigma^2)) - log(sigma * sqrt(2*pi))

        For the maximum likelihood estimator we use the negative of this sum:

        .. code-block:: c

            -sum_n(log(PDF)).
        """
        super(OffsetGaussianEvaluationModel, self).__init__(
            'OffsetGaussianNoise',
            'offsetGaussianNoiseModel',
            (FreeParameter(CLDataType.from_string('mot_float_type'), 'sigma', True, 1, 0, 'INFINITY',
                           parameter_transform=ClampTransform()),), ())

    def get_objective_function(self, fname, inst_per_problem, eval_fname, obs_fname, param_listing):
        """Get the Offset Gaussian objective function.

        This omits the constant terms for speed reasons. Omitted terms are:

         .. code-block:: c

            (+ log(OffsetGaussianNoise_sigma * sqrt(2 * M_PI)))
        """
        return '''
            double ''' + fname + '''(const void* const data, mot_float_type* const x){
                ''' + param_listing + '''
                double sum = 0.0;
                for(int i = 0; i < ''' + str(inst_per_problem) + '''; i++){
                    sum += pown(''' + obs_fname + '''(data, i) -
                                sqrt(pown(''' + eval_fname + '''(data, x, i), 2) +
                                    (OffsetGaussianNoise_sigma * OffsetGaussianNoise_sigma)), 2);
                }
                return sum / (2 * pown(OffsetGaussianNoise_sigma, 2));
            }
        '''

    def get_objective_list_function(self, fname, inst_per_problem, eval_fname, obs_fname, param_listing):
        return '''
            void ''' + fname + '''(const void* const data, mot_float_type* const x, mot_float_type* result){
                ''' + param_listing + '''
                for(int i = 0; i < ''' + str(inst_per_problem) + '''; i++){
                    result[i] = ''' + obs_fname + '''(data, i) -
                                    sqrt(pown(''' + eval_fname + '''(data, x, i), 2)
                                            + (OffsetGaussianNoise_sigma * OffsetGaussianNoise_sigma));
                }
            }
        '''

    def get_log_likelihood_function(self, fname, inst_per_problem, eval_fname, obs_fname, param_listing,
                                    full_likelihood=True):
        return '''
            double ''' + fname + '''(const void* const data, mot_float_type* const x){
                ''' + param_listing + '''
                double sum = 0.0;
                for(int i = 0; i < ''' + str(inst_per_problem) + '''; i++){
                    sum += pown(''' + obs_fname + '''(data, i) -
                                sqrt(pown(''' + eval_fname + '''(data, x, i), 2) +
                                    (OffsetGaussianNoise_sigma * OffsetGaussianNoise_sigma)), 2);
                }
                return - sum / (2 * pown(OffsetGaussianNoise_sigma, 2))
                    ''' + ('+' + str(inst_per_problem) + ' * log(OffsetGaussianNoise_sigma * sqrt(2 * M_PI))'
                           if full_likelihood else '') + ''';
            }
        '''


class RicianEvaluationModel(EvaluationModel):

    def __init__(self):
        """This uses the log of the Rice PDF for the maximum likelihood estimator and for the log likelihood.

        The PDF is defined as:

        .. code-block:: c

            PDF = (observation/sigma^2)
                    * exp(-(observation^2 + evaluation^2) / (2 * sigma^2))
                    * bessel_i0((observation * evaluation) / sigma^2)

        Where where ``bessel_i0(z)`` is the modified Bessel function of the first kind with order zero. To have the
        joined probability over all instances one would have to take the product over all n instances:

        .. code-block:: c

            product(PDF)

        Instead of taking the product of this PDF over all instances we take the sum of the log of the PDF:

        .. code-block:: c

            sum(log(PDF))

        Where the log of the PDF is given by:

        .. code-block:: c

            log(PDF) = log(observation/sigma^2)
                        - (observation^2 + evaluation^2) / (2 * sigma^2)
                        + log(bessel_i0((observation * evaluation) / sigma^2))

        For the maximum likelihood estimator we use the negative of this sum:

        .. code-block:: c

            -sum(log(PDF)).
        """
        super(RicianEvaluationModel, self).__init__(
            'RicianNoise',
            'ricianNoiseModel',
            (FreeParameter(CLDataType.from_string('mot_float_type'), 'sigma', True, 1, 0, 'INFINITY',
                           parameter_transform=ClampTransform()),),
            (Bessel(),))

    def get_objective_function(self, fname, inst_per_problem, eval_fname, obs_fname, param_listing):
        """Get the Rician objective function.

        This omits the constant terms for speed reasons. Omitted terms are:

         .. code-block:: c

            + log(observation / (RicianNoise_sigma * RicianNoise_sigma))
            - ((observation * observation) / (2 * (RicianNoise_sigma * RicianNoise_sigma)))

        """
        return '''
            double ''' + fname + '''(const void* const data, mot_float_type* const x){
                ''' + param_listing + '''
                double sum = 0.0;
                double observation;
                double evaluation;
                for(int i = 0; i < ''' + str(inst_per_problem) + '''; i++){
                    observation = (double)''' + obs_fname + '''(data, i);
                    evaluation = (double)''' + eval_fname + '''(data, x, i);

                    sum +=  - ((evaluation * evaluation) / (2 * RicianNoise_sigma * RicianNoise_sigma))
                            + log_bessel_i0((observation * evaluation) / (RicianNoise_sigma * RicianNoise_sigma));
                }
                return -sum;
            }
        '''

    def get_objective_list_function(self, fname, inst_per_problem, eval_fname, obs_fname, param_listing):
        return '''
            void ''' + fname + '''(const void* const data, mot_float_type* const x, mot_float_type* result){
                ''' + param_listing + '''
                double observation;
                double evaluation;
                for(int i = 0; i < ''' + str(inst_per_problem) + '''; i++){
                    observation = (double)''' + obs_fname + '''(data, i);
                    evaluation = (double)''' + eval_fname + '''(data, x, i);

                    result[i] = - ((evaluation * evaluation) / (2 * RicianNoise_sigma * RicianNoise_sigma))
                                + log_bessel_i0((observation * evaluation) / (RicianNoise_sigma * RicianNoise_sigma));
                }
            }
        '''

    def get_log_likelihood_function(self, fname, inst_per_problem, eval_fname, obs_fname, param_listing,
                                    full_likelihood=True):
        return '''
            double ''' + fname + '''(const void* const data, const mot_float_type* const x){
                ''' + param_listing + '''
                double sum = 0.0;
                double observation;
                double evaluation;
                for(int i = 0; i < ''' + str(inst_per_problem) + '''; i++){
                    observation = (double)''' + obs_fname + '''(data, i);
                    evaluation = (double)''' + eval_fname + '''(data, x, i);

                    sum += log(observation / (RicianNoise_sigma * RicianNoise_sigma))
                            - ((observation * observation) / (2 * RicianNoise_sigma * RicianNoise_sigma))
                            - ((evaluation * evaluation) / (2 * RicianNoise_sigma * RicianNoise_sigma))
                            + log_bessel_i0((observation * evaluation) / (RicianNoise_sigma * RicianNoise_sigma));
                }
                return sum;
            }
        '''



import pyopencl as cl
import numpy as np

from mot.utils import is_scalar
from ...cl_routines.optimizing.base import AbstractParallelOptimizer, AbstractParallelOptimizerWorker

__author__ = 'Robbert Harms'
__date__ = "2014-02-05"
__license__ = "LGPL v3"
__maintainer__ = "Robbert Harms"
__email__ = "robbert.harms@maastrichtuniversity.nl"


class GridGenerator(object):

    def get_grid(self, model, grid_size):
        """Get the parameter grid we use to evaluate the model against.

        Args:
            model: the model for which to generate or return a parameter grid
            grid_size: an indication of the desired grid size, the grid may be larger

        Returns:
            ndarray: a grid of parameters with for every parameter (column) an starting point (row). That is,
                every row contains a unique starting point.
        """
        return NotImplementedError


class GivenGrid(GridGenerator):

    def __init__(self, grid):
        """Use the given grid for the computations.

        Args:
            grid (ndarray): the grid we use for the computations. The second dimension should match the number
                of parameters in the model.
        """
        self._grid = grid

    def get_grid(self, model, grid_size):
        return self._grid


class LinearSpacedGrid(GridGenerator):

    def __init__(self, grid_nmr_steps=None):
        """Creates a linear spaced grid for the grid search.

        The steps per dimension can be controlled using the parameter ``grid_nmr_steps``. If this is a scalar
        we use that for all dimensions, if it is a list we use the steps given for each dimension separately.

        Setting a grid step to 1 or lower disables that dimension from the grid search and the average
        initial parameter value is returned instead.

        If the grid number of steps is not given we use the 'grid_size' argument in the method 'get_grid' for
        generating the grid (typically set to the patience of the optimizer). The total grid size in that case is then
        ``grid_size * nmr_parameters``.

        Args:
            grid_nmr_steps (int or list): the number of steps for the generated grid. If set to None we
                use the argument 'grid_size' of the method 'get_grid' for generating the grid.
        """
        self._grid_nmr_steps = grid_nmr_steps

    def get_grid(self, model, grid_size):
        return self._generate_grid(model, grid_size)

    def _generate_grid(self, model, grid_size):
        """Generate a grid for the grid search

        Returns:
            ndarray: a nxm array where n is the number of grid points and m is the number of parameters
        """
        spacings = self._get_grid_spacing(model, grid_size)

        np_dtype = np.float32
        if model.double_precision:
            np_dtype = np.float64

        lower_bounds = model.get_lower_bounds()
        upper_bounds = model.get_upper_bounds()
        default_params = np.mean(model.get_initial_parameters(), axis=0)

        result = np.zeros((1, model.get_nmr_estimable_parameters()), np_dtype)

        repeat_mult = 1
        for index, spacing in enumerate(spacings):
            if spacing <= 1:
                spacing = 1
                values = default_params[index]
            else:
                values = np.linspace(np.min(lower_bounds[index]), np.max(upper_bounds[index]), spacing)

            result = np.tile(result, (spacing, 1))
            result[:, index] = np.repeat(values, repeat_mult)

            repeat_mult *= spacing

        return result

    def _get_grid_spacing(self, model, grid_size):
        """This returns the grid spacings used to generate a grid automatically.

        Returns:
            list: the grid spacings with one integer for every dimension.
        """
        if self._grid_nmr_steps is None:
            return [np.round(grid_size ** (1.0/model.get_nmr_estimable_parameters()))] \
                        * model.get_nmr_estimable_parameters()
        elif is_scalar(self._grid_nmr_steps):
            return [self._grid_nmr_steps] * model.get_nmr_estimable_parameters()

        if len(self._grid_nmr_steps) != model.get_nmr_estimable_parameters():
            raise ValueError('The length of the given grid ({}) steps does '
                             'not equal the number of parameters ({}).'.format(self._grid_nmr_steps,
                                                                               model.get_nmr_estimable_parameters()))

        return self._grid_nmr_steps


class UniformRandomGrid(GridGenerator):

    def __init__(self):
        """Creates an uniform random grid for the grid search."""

    def get_grid(self, model, grid_size):
        np_dtype = np.float32
        if model.double_precision:
            np_dtype = np.float64

        result = np.zeros((grid_size, model.get_nmr_estimable_parameters()), np_dtype)

        lower_bounds = model.get_lower_bounds()
        upper_bounds = model.get_upper_bounds()

        for param_ind in range(len(lower_bounds)):
            result[:, param_ind] = np.random.uniform(np.min(lower_bounds[param_ind]),
                                                     np.max(upper_bounds[param_ind]), grid_size)

        return result


class GaussianRandomGrid(GridGenerator):

    def __init__(self):
        """Generates a random parameter grid using a normal distribution around the current initial value.

        This uses the mean of the initial value as the mean of the distribution and half the difference
        between the lower and upper bounds as the std.
        """

    def get_grid(self, model, grid_size):
        np_dtype = np.float32
        if model.double_precision:
            np_dtype = np.float64

        result = np.zeros((grid_size, model.get_nmr_estimable_parameters()), np_dtype)

        lower_bounds = model.get_lower_bounds()
        upper_bounds = model.get_upper_bounds()
        default_params = np.mean(model.get_initial_parameters(), axis=0)

        for param_ind in range(len(lower_bounds)):
            result[:, param_ind] = np.random.normal(
                default_params[param_ind],
                (np.max(upper_bounds[param_ind]) - np.min(lower_bounds[param_ind])) / 2.0,
                grid_size)

        return result


class GridSearch(AbstractParallelOptimizer):

    default_patience = 250

    def __init__(self, grid_generator=None, patience=None, **kwargs):
        """A grid search routine that searches an grid for a good solution.

        This optimization routine evaluates the model at every point in a grid (the same points for every
        problem instance) and retunrs per problem instance the parameters with the lowest function value.

        Please note that the current starting point is always evaluated as well and can be returned as optimum for a
        problem instance.

        For generating the grid we use an instance of an :class:`ParameterGrid`.

        Args:
            grid_generator (:ref:`ParameterGrid`): the parameter grid to use
        """
        self._grid_generator = grid_generator
        patience = patience or self.default_patience
        super(GridSearch, self).__init__(patience=patience, **kwargs)

    def _get_worker_generator(self, *args):
        model = args[1]
        nmr_params = args[3]

        np_dtype = np.float32
        if model.double_precision:
            np_dtype = np.float64

        grid = np.require(self._grid_generator.get_grid(
            model, self.patience * nmr_params), np_dtype, requirements=['C', 'A', 'O'])

        if grid.shape[1] != model.get_nmr_estimable_parameters():
            raise ValueError('The shape of the generated grid is not compatible with the given model.')

        return lambda cl_environment: GridSearchWorker(cl_environment, *args, grid=grid)


class GridSearchWorker(AbstractParallelOptimizerWorker):

    def __init__(self, *args, **kwargs):
        self._grid = kwargs.pop('grid')
        super(GridSearchWorker, self).__init__(*args, **kwargs)

    def _create_buffers(self):
        buffers = super(GridSearchWorker, self)._create_buffers()

        parameters_buffer = cl.Buffer(self._cl_run_context.context,
                                      cl.mem_flags.READ_WRITE | cl.mem_flags.USE_HOST_PTR,
                                      hostbuf=self._grid)
        buffers[0].append(parameters_buffer)
        return buffers

    def _get_kernel_param_names(self):
        kernel_param_names = super(GridSearchWorker, self)._get_kernel_param_names()
        kernel_param_names.append('global mot_float_type* grid')
        return kernel_param_names

    def _get_optimizer_call_args(self):
        call_args = super(GridSearchWorker, self)._get_optimizer_call_args()
        call_args.append('grid')
        return call_args

    def _get_optimization_function(self):
        nmr_params = self._nmr_params

        kernel_source = ''
        kernel_source += self._model.get_parameter_encode_function('encodeParameters') + "\n"

        kernel_source += '''
            int grid_search(mot_float_type* model_parameters, const void* const data, global mot_float_type* grid){

                int param_ind;

                mot_float_type best_params_so_far[''' + str(nmr_params) + '''];
                for (param_ind = 0; param_ind < ''' + str(nmr_params) + '''; param_ind++){
                    best_params_so_far[param_ind] = model_parameters[param_ind];
                }

                mot_float_type lowest_error = evaluate(model_parameters, data);
                mot_float_type error = 0.0;

                for(int i = 0; i < ''' + str(self._grid.shape[0]) + '''; i++){

                    for (param_ind = 0; param_ind < ''' + str(nmr_params) + '''; param_ind++){
                        model_parameters[param_ind] = grid[i * ''' + str(nmr_params) + ''' + param_ind];
                    }

                    encodeParameters(model_parameters, (void*)&data);
                    error = evaluate(model_parameters, data);

                    if(error < lowest_error){
                        lowest_error = error;

                        for (param_ind = 0; param_ind < ''' + str(nmr_params) + '''; param_ind++){
                            best_params_so_far[param_ind] = model_parameters[param_ind];
                        }
                    }
                }

                for (param_ind = 0; param_ind < ''' + str(nmr_params) + '''; param_ind++){
                    model_parameters[param_ind] = best_params_so_far[param_ind];
                }

                return 0;
            }
        '''
        return kernel_source

    def _get_optimizer_call_name(self):
        return 'grid_search'



import os
from pkg_resources import resource_filename
import pyopencl as cl
from mot.utils import get_float_type_def
from .base import AbstractParallelOptimizer, AbstractParallelOptimizerWorker

__author__ = 'Robbert Harms'
__date__ = "2014-02-05"
__license__ = "LGPL v3"
__maintainer__ = "Robbert Harms"
__email__ = "robbert.harms@maastrichtuniversity.nl"


class LevenbergMarquardt(AbstractParallelOptimizer):

    default_patience = 250

    def __init__(self, patience=None, **kwargs):
        """Use the Levenberg-Marquardt method to calculate the optimimum.

        Args:
            patience (int):
                Used to set the maximum number of iterations to patience*(number_of_parameters+1)
        """
        patience = patience or self.default_patience
        super(LevenbergMarquardt, self).__init__(patience=patience, **kwargs)

    def _get_worker_generator(self, *args):
        return lambda cl_environment: LevenbergMarquardtWorker(cl_environment, *args)


class LevenbergMarquardtWorker(AbstractParallelOptimizerWorker):

    def __init__(self, *args, **kwargs):
        super(LevenbergMarquardtWorker, self).__init__(*args, **kwargs)

        if self._model.get_nmr_inst_per_problem() < self._nmr_params:
            raise ValueError('The number of instances per problem must be greater than the number of parameters')

    def _create_buffers(self):
        all_buffers, parameters_buffer, return_code_buffer = super(LevenbergMarquardtWorker, self)._create_buffers()

        fjac_items = self._nmr_params * self._model.get_nmr_inst_per_problem() * self._starting_points.shape[0]
        fjac_buffer_size = fjac_items * self._starting_points.dtype.itemsize
        fjac_buffer = cl.Buffer(self._cl_run_context.context, cl.mem_flags.READ_WRITE, size=fjac_buffer_size)
        all_buffers.append(fjac_buffer)

        return all_buffers, parameters_buffer, return_code_buffer

    def _get_kernel_source(self):
        """Overwrite the default kernel source generation.

        We need to add a few extra buffers to the kernel and the lmmin call. This makes the optimizer
        more robust when using datasets with a high number of observations per problem instance.

        If we inline these buffers the CL compiler may throw a 'out of resources' error since it can fail to find
        a contiguous block of memory large enough for any of these buffers. We circumvent that now by creating them
        in global memory space.
        """
        nmr_params = self._nmr_params

        kernel_param_names = ['global mot_float_type* params',
                              'global char* return_codes']
        kernel_param_names.extend(self._model.get_kernel_param_names(self._cl_environment.device))
        kernel_param_names.append('global mot_float_type* fjac_all')

        optimizer_call_args = 'x, (const void*) &data'

        kernel_source = ''
        kernel_source += get_float_type_def(self._double_precision)
        kernel_source += str(self._model.get_kernel_data_struct(self._cl_environment.device))

        if self._use_param_codec:
            kernel_source += self._model.get_parameter_decode_function('decodeParameters') + "\n"

        kernel_source += self._get_optimizer_cl_code()
        kernel_source += '''
            __kernel void minimize(
                ''' + ",\n".join(kernel_param_names) + '''
                ){
                    int gid = get_global_id(0);

                    mot_float_type x[''' + str(nmr_params) + '''];
                    for(int i = 0; i < ''' + str(nmr_params) + '''; i++){
                        x[i] = params[gid * ''' + str(nmr_params) + ''' + i];
                    }

                    global mot_float_type* fjac = fjac_all + gid * ''' \
                         + str(self._nmr_params * self._model.get_nmr_inst_per_problem()) + ''';

                    ''' + self._model.get_kernel_data_struct_initialization(self._cl_environment.device, 'data') + '''
                    return_codes[gid] = (char) ''' + self._get_optimizer_call_name() + '''(''' \
                         + optimizer_call_args + ''', fjac);

                    ''' + ('decodeParameters((void*)&data, x);' if self._use_param_codec else '') + '''

                    for(int i = 0; i < ''' + str(nmr_params) + '''; i++){
                        params[gid * ''' + str(nmr_params) + ''' + i] = x[i];
                    }
                }
        '''
        return kernel_source

    def _get_evaluate_function(self):
        """Get the CL code for the evaluation function. This is called from _get_optimizer_cl_code.

        Implementing optimizers can change this if desired.

        Returns:
            str: the evaluation function.
        """
        kernel_source = ''
        kernel_source += self._model.get_objective_list_function('calculateObjectiveList')
        if self._use_param_codec:
            kernel_source += '''
                void evaluate(mot_float_type* x, const void* data, mot_float_type* result){
                    mot_float_type x_model[''' + str(self._nmr_params) + '''];
                    for(int i = 0; i < ''' + str(self._nmr_params) + '''; i++){
                        x_model[i] = x[i];
                    }
                    decodeParameters(data, x_model);
                    calculateObjectiveList(data, x_model, result);
                }
            '''
        else:
            kernel_source += '''
                void evaluate(mot_float_type* x, const void* data, mot_float_type* result){
                    calculateObjectiveList(data, x, result);
                }
            '''
        return kernel_source

    def _get_optimization_function(self):
        params = {'NMR_PARAMS': self._nmr_params,
                  'PATIENCE': self._parent_optimizer.patience,
                  'NMR_INST_PER_PROBLEM': self._model.get_nmr_inst_per_problem()}

        optimizer_settings = self._optimizer_settings or {}
        option_defaults = {'step_bound': 100.0, 'scale_diag': 1}
        option_converters = {'scale_diag': lambda val: int(bool(val))}

        for option, default in option_defaults.items():
            v = optimizer_settings.get(option, default)
            if option in option_converters:
                v = option_converters[option](v)
            params.update({option.upper(): v})

        body = open(os.path.abspath(resource_filename('mot', 'data/opencl/lmmin.pcl')), 'r').read()
        if params:
            body = body % params
        return body

    def _get_optimizer_call_name(self):
        return 'lmmin'



import numpy as np

from mot.cl_data_type import CLDataType
from mot.cl_routines.mapping.calc_dependent_params import CalculateDependentParameters
from mot.model_building.cl_functions.parameters import CurrentObservationParam, StaticMapParameter, ProtocolParameter, \
    ModelDataParameter, FreeParameter
from mot.model_building.data_adapter import SimpleDataAdapter
from mot.model_building.parameter_functions.dependencies import SimpleAssignment
from mot.model_interfaces import OptimizeModelInterface, SampleModelInterface
from mot.utils import TopologicalSort, is_scalar, all_elements_equal, get_single_value

__author__ = 'Robbert Harms'
__date__ = "2014-03-14"
__license__ = "LGPL v3"
__maintainer__ = "Robbert Harms"
__email__ = "robbert.harms@maastrichtuniversity.nl"


class OptimizeModelBuilder(OptimizeModelInterface):

    def __init__(self, name, model_tree, evaluation_model, signal_noise_model=None, problem_data=None):
        """Create a new model builder that can construct an optimization model using parts.

        Args:
            name (str): the name of the model
            model_tree (mot.model_building.trees.CompartmentModelTree): the model tree object
            evaluation_model (mot.model_building.evaluation_models.EvaluationModel): the evaluation model to
                use for the resulting complete model
            signal_noise_model (mot.model_building.signal_noise_models.SignalNoiseModel): the optional signal
                noise model to use to add noise to the model prediction
            problem_data (ProblemData): the problem data object

        Attributes:
            problems_to_analyze (list): the list with problems we want to analyze. Suppose we have a few thousands
                problems defined in this model, but we want to run the optimization only on a few problems. By setting
                this attribute to a list of problem indices, only those problems will be analyzed.
            use_parameter_transformations (boolean): set to False to disable the parameter transformations.
                This basically sets the encode and decode functions to the identity function.
        """
        super(OptimizeModelBuilder, self).__init__()

        self.use_parameter_transformations = True

        self._dependency_store = _DependencyStore()
        self._model_functions_info = _ModelFunctionsInformation(self._dependency_store,
                                                                model_tree, evaluation_model, signal_noise_model)

        self._name = name
        self._double_precision = False
        self._model_tree = model_tree
        self._evaluation_model = evaluation_model
        self._signal_noise_model = signal_noise_model
        self._parameters_dot_to_bar = {}
        self._post_optimization_modifiers = []
        self.problems_to_analyze = None

        # The values to use for the parameters, this is the place subclasses should use for their information.
        # At init this is filled with the values as defined in the parameters.
        self._parameter_values = {'{}.{}'.format(m.name, p.name): p.value for m, p in
                                  self._model_functions_info.get_free_parameters_list() +
                                  self._model_functions_info.get_static_parameters_list()}

        self._lower_bounds = {'{}.{}'.format(m.name, p.name): p.lower_bound for m, p in
                              self._model_functions_info.get_free_parameters_list()}

        self._upper_bounds = {'{}.{}'.format(m.name, p.name): p.upper_bound for m, p in
                              self._model_functions_info.get_free_parameters_list()}

        self._problem_data = None
        if problem_data:
            self.set_problem_data(problem_data)

        for m, p in self._model_functions_info.get_model_parameter_list():
            self._parameters_dot_to_bar.update({m.name + '.' + p.name: m.name + '_' + p.name})

        self._set_default_dependencies()

    @property
    def name(self):
        """See super class OptimizeModelInterface for details"""
        return self._name

    @property
    def double_precision(self):
        """See super class OptimizeModelInterface for details"""
        return self._double_precision

    @double_precision.setter
    def double_precision(self, value):
        """Set the value for double_precision.

        Args:
            value (boolean): if we would like to do the computations in double of single floating point type.
        """
        self._double_precision = value

    def fix(self, model_param_name, value):
        """Fix the given model.param to the given value.

        Args:
            model_param_name (string): A model.param name like 'Ball.d'
            value (scalar or vector): The value to fix the given parameter to

        Returns:
            Returns self for chainability
        """
        self._parameter_values[model_param_name] = value
        self._model_functions_info.set_fixed_to_value(model_param_name, True)
        return self

    def init(self, model_param_name, value):
        """Init the given model.param to the given value.

        Args:
            model_param_name (string): A model.param name like 'Ball.d'
            value (scalar or vector): The value to initialize the given parameter to

        Returns:
            Returns self for chainability
        """
        self._parameter_values[model_param_name] = value
        return self

    def set_lower_bound(self, model_param_name, value):
        """Set the lower bound for the given parameter to the given value.

        Args:
            model_param_name (string): A model.param name like 'Ball.d'
            value (scalar or vector): The value to set the lower bounds to

        Returns:
            Returns self for chainability
        """
        self._lower_bounds[model_param_name] = value
        return self

    def set_upper_bound(self, model_param_name, value):
        """Set the upper bound for the given parameter to the given value.

        Args:
            model_param_name (string): A model.param name like 'Ball.d'
            value (scalar or vector): The value to set the upper bounds to

        Returns:
            Returns self for chainability
        """
        self._upper_bounds[model_param_name] = value
        return self

    def set_parameter_transform(self, model_param_name, value):
        """Set the parameter transform for the given parameter to the given transformation.

        Args:
            model_param_name (string): A model.param name like 'Ball.d'
            value (AbstractTransformation): The parameter transform to use

        Returns:
            Returns self for chainability
        """
        m, p = self._model_functions_info.get_model_parameter_by_name(model_param_name)
        p.parameter_transform = value
        return self

    def unfix(self, model_param_name):
        """Unfix the given model.param

        Args:
            model_param_name (string): A model.param name like 'Ball.d'

        Returns:
            Returns self for chainability
        """
        self._model_functions_info.set_fixed_to_value(model_param_name, False)
        return self

    def has_parameter(self, model_param_name):
        """Check to see if the given parameter is defined in this model.

        Args:
            model_param_name (string): A model.param name like 'Ball.d'

        Returns:
            boolean: true if the parameter is defined in this model, false otherwise.
        """
        return self._model_functions_info.has_parameter(model_param_name)

    def set_problem_data(self, problem_data):
        """Set the problem data this model will deal with.

        This will also call the function set_noise_level_std() with the noise_std from the new problem data.

        Args:
            problem_data (ProblemData):
                The container for the problem data we will use for this model.

        Returns:
            Returns self for chainability
        """
        self._problem_data = problem_data
        if self._problem_data.noise_std is not None:
            self._parameter_values['{}.{}'.format(
                self._evaluation_model.name,
                self._evaluation_model.get_noise_std_param_name())] = self._problem_data.noise_std
        return self

    def add_parameter_dependency(self, parameter_name, dependency):
        """Adds a dependency rule to this model. The dependency is supposed to be a ParameterDependency object.

        The dependencies are executed in the same order as they were added to this model.

        Args:
            parameter_name (String): The parameter on which the dependency is applied
            dependency (ParameterDependency): The dependency rule, an ParameterDependency object
        """
        if parameter_name not in self._parameters_dot_to_bar:
            raise ParameterNameException("The parameter name \"{}\" can not be "
                                         "found in the model listing.".format(parameter_name))
        self._dependency_store.set_dependency(parameter_name, dependency)
        return self

    def add_parameter_dependencies(self, dependencies):
        """Adds a list of dependency objects. The order of the dependencies matter (order of parameter instantiation).

        Args:
            dependencies: the dependency rules, an tuple list like ((name, ParameterDependency object),)
        """
        for name, d in dependencies:
            self.add_parameter_dependency(name, d)
        return self

    def add_post_optimization_modifier(self, model_param_name, mod_routine):
        """Add a modification function that can update the results of model optimization.

        The mod routine should be a function accepting a dictionary as input and should return a single map of
        the same dimension as the maps in the dictionary. The idea is that the mod_routine function gets the
        result dictionary from the optimization routine and calculates a new map.

        This map is returned and the dictionary is updated with the returned map as value and the here given
        model_param_name as key.

        It is possible to add more than one modifier function. In that case, they are called in the order they
        were appended to this model.

        Args:
            model_param_name (str): the parameter to which to add the modification routine
            mod_routine (python function): the callback function to apply on the results of the referenced parameter.
        """
        self._post_optimization_modifiers.append((model_param_name, mod_routine))
        return self

    def add_post_optimization_modifiers(self, modifiers):
        """Add a list of modifier functions.

        The same as add_post_optimization_modifier() except that it accepts a list of lists. Every element in the list
        should be a tuple like (model_param_name, mod_routine)

        Args:
            modifiers (tuple or list): The list of modifiers to add (in order).

        """
        self._post_optimization_modifiers.extend(modifiers)

    def get_required_protocol_names(self):
        """Get a list with the constant data names that are needed for this model to work.

        For example, an implementing diffusion MRI model might require the presence of the protocol parameter
        'g' and 'b'. This function should then return ('g', 'b').

        Returns:
            list: A list of columns names that need to be present in the protocol
        """
        return list(set([p.name for m, p in self._model_functions_info.get_model_parameter_list() if
                         isinstance(p, ProtocolParameter)]))

    def get_optimization_output_param_names(self):
        """See super class for details"""
        items = ['{}.{}'.format(m.name, p.name) for m, p in self._model_functions_info.get_free_parameters_list()]
        items.extend(name for name, _ in self._post_optimization_modifiers)
        return items

    def get_optimized_param_names(self):
        """See super class for details"""
        return ['{}.{}'.format(m.name, p.name) for m, p in
                self._model_functions_info.get_estimable_parameters_list()]

    def get_nmr_problems(self):
        """See super class for details"""
        if self.problems_to_analyze is None:
            return self._problem_data.get_nmr_problems()
        return len(self.problems_to_analyze)

    def get_nmr_inst_per_problem(self):
        """See super class for details"""
        return self._problem_data.get_nmr_inst_per_problem()

    def get_nmr_estimable_parameters(self):
        """See super class for details"""
        return len(self.get_optimized_param_names())

    def get_data(self):
        """See super class for details"""
        data = []
        for data_dict in [self._get_variable_data(), self._get_protocol_data(), self._get_static_data()]:
            for el in data_dict.values():
                data.append(el.get_opencl_data())
        return data

    def get_kernel_data_struct(self, device):
        """See super class for details"""
        return self._get_all_kernel_source_items(device)['data_struct']

    def get_kernel_param_names(self, device):
        """See super class for details"""
        return self._get_all_kernel_source_items(device)['kernel_param_names']

    def get_kernel_data_struct_initialization(self, device, variable_name):
        """See super class for details"""
        data_struct_init = self._get_all_kernel_source_items(device)['data_struct_init']
        struct_code = '0'
        if data_struct_init:
            struct_code = ', '.join(data_struct_init)
        return self.get_kernel_data_struct_type() + ' ' + variable_name + ' = {' + struct_code + '};'

    def get_kernel_data_struct_type(self):
        """Get the CL type of the kernel datastruct.

        Returns:
            str: the CL type of the data struct
        """
        return '_model_data'

    def get_parameter_decode_function(self, fname='decodeParameters'):
        func = '''
            void ''' + fname + '''(const void* data_void, mot_float_type* x){
        '''
        if self.use_parameter_transformations:
            func += self.get_kernel_data_struct_type() + \
                    '* data = (' + self.get_kernel_data_struct_type() + '*)data_void;'
            for d in self._get_parameter_transformations()[1]:
                func += "\n" + "\t" * 4 + d.format('x')
        return func + '''
            }
        '''

    def get_parameter_encode_function(self, fname='encodeParameters'):
        func = '''
            void ''' + fname + '''(const void* data_void, mot_float_type* x){
        '''
        if self.use_parameter_transformations:
            func += self.get_kernel_data_struct_type() + \
                    '* data = (' + self.get_kernel_data_struct_type() + '*)data_void;'
            for d in self._get_parameter_transformations()[0]:
                func += "\n" + "\t" * 4 + d.format('x')

        return func + '''
            }
        '''

    def get_initial_parameters(self, results_dict=None):
        """When overriding this function, please note that it should adhere to the attribute problems_to_analyze.

        Args:
            results_dict (dict): the initialization settings for the specific parameters.
                The number of items per dictionary item should match the number of problems to analyze.
        """
        np_dtype = np.float32
        if self.double_precision:
            np_dtype = np.float64

        starting_points = []
        for m, p in self._model_functions_info.get_estimable_parameters_list():
            param_name = '{}.{}'.format(m.name, p.name)
            value = self._parameter_values[param_name]

            if results_dict and param_name in results_dict:
                starting_points.append(results_dict[m.name + '.' + p.name])
            elif is_scalar(value):
                starting_points.append(np.full((self.get_nmr_problems(), 1), value, dtype=np_dtype))
            else:
                if len(value.shape) < 2:
                    value = np.transpose(np.asarray([value]))
                elif value.shape[1] > value.shape[0]:
                    value = np.transpose(value)
                else:
                    value = value

                if self.problems_to_analyze is None:
                    starting_points.append(value)
                else:
                    starting_points.append(value[self.problems_to_analyze, ...])

        starting_points = np.concatenate([np.transpose(np.array([s]))
                                          if len(s.shape) < 2 else s for s in starting_points], axis=1)

        data_adapter = SimpleDataAdapter(starting_points, CLDataType.from_string('mot_float_type'),
                                         self._get_mot_float_type())
        return data_adapter.get_opencl_data()

    def get_lower_bounds(self):
        """See super class for details"""
        return list(self._lower_bounds['{}.{}'.format(m.name, p.name)] for m, p in
                    self._model_functions_info.get_estimable_parameters_list())

    def get_upper_bounds(self):
        """See super class for details"""
        return list(self._upper_bounds['{}.{}'.format(m.name, p.name)] for m, p in
                    self._model_functions_info.get_estimable_parameters_list())

    def set_initial_parameters(self, initial_params):
        """Update the initial parameters for this model by the given values.

        This only affects free non-fixed parameters.

        Args:
            initial_params (dict): a dictionary containing as keys full parameter names (<model>.<param>) and as values
                numbers or arrays to be used as starting point
        """
        for m, p in self._model_functions_info.get_estimable_parameters_list():
            param_name = '{}.{}'.format(m.name, p.name)

            if param_name in initial_params:
                self._parameter_values[param_name] = initial_params[param_name]

        return self

    def get_final_parameter_transformations(self, func_name='applyFinalParameterTransformations'):
        """Get the transformations that must be applied at the end of an optimization (or sampling) routine.

        These transformations must contain all parameter dependencies, as such that all transformation happening in the
        model function which do not happen in the codec must also go here.

        Args:
            func_name (str): the CL function name of the returned function

        Returns:
            str: A function of the kind:
                void finalParameterTransformations(const void* data, mot_float_type* x)
                Which is called for every voxel and must in place edit the x variable.
        """
        transform_needed = any(dp.has_side_effects or not dp.fixed for dp in
                               self._dependency_store.dependencies.values())

        if not self._dependency_store.has_dependencies() or not transform_needed:
            return None

        param_exclude_list = [m.name + '_' + p.name for (m, p)
                              in self._model_functions_info.get_non_model_tree_param_listing()]
        param_lists = self._get_parameter_type_lists()
        depend_param_listing = self._get_dependent_parameters_listing(param_lists['dependent'])

        for m, p in param_lists['fixed'] + param_lists['protocol']:
            if (m.name + '_' + p.name) not in depend_param_listing:
                param_exclude_list.append(m.name + '_' + p.name)

        param_listing = self._get_parameters_listing(exclude_list=param_exclude_list)

        func = "\n\t\t\t" + 'void ' + func_name + '(const void* const void_data, mot_float_type* const x){' + "\n"
        func += "\n\t\t\t\t" + 'const ' + self.get_kernel_data_struct_type() + \
                '* data = (' + self.get_kernel_data_struct_type() + '*)void_data;'
        func += param_listing + "\n"

        for i, (m, p) in enumerate(self._get_parameter_type_lists()['estimable']):
            if not self._is_non_model_tree_model(m):
                func += "\t"*4 + 'x[' + str(i) + '] = ' + m.name + '_' + p.name + ';' + "\n"
        func += "\t\t\t" + '}' + "\n"
        return func

    def get_observation_return_function(self, func_name='getObservation'):
        if self._problem_data.observations.shape[1] < 2:
            return '''
                mot_float_type ''' + func_name + '''(const void* const data, const int observation_index){
                    return ((''' + self.get_kernel_data_struct_type() + '''*)data)->var_data_observations;
                }
            '''

        return '''
            mot_float_type ''' + func_name + '''(const void* const data, const int observation_index){
                return ((''' + \
                    self.get_kernel_data_struct_type() + '''*)data)->var_data_observations[observation_index];
            }
        '''

    def get_model_eval_function(self, func_name='evaluateModel'):
        noise_func_name = func_name + '_signalNoiseModel'
        func = self._get_model_functions_cl_code(noise_func_name)

        pre_model_function = self._get_pre_model_expression_eval_function()
        if pre_model_function:
            func += pre_model_function

        func += '''
            mot_float_type ''' + func_name + \
                '(const void* const void_data, const mot_float_type* const x, const int observation_index){' + "\n"
        func += self.get_kernel_data_struct_type() + '* data = (' + self.get_kernel_data_struct_type() + '*)void_data;'

        func += self._get_parameters_listing(
            exclude_list=[m.name + '_' + p.name for (m, p) in
                          self._model_functions_info.get_non_model_tree_param_listing()])

        if self._signal_noise_model:
            noise_params_listing = ''
            for p in self._signal_noise_model.get_free_parameters():
                noise_params_listing += "\t" * 4 + self._get_param_listing_for_param(self._signal_noise_model, p)
            func += "\n"
            func += noise_params_listing

        pre_model_code = self._get_pre_model_expression_eval_code()
        if pre_model_code:
            func += self._get_pre_model_expression_eval_code()

        func += "\n" + "\t"*4 + 'return ' + str(self._construct_model_expression(noise_func_name))
        func += "\n\t\t\t}"
        return func

    def get_objective_function(self, func_name="calculateObjective"):
        inst_per_problem = self.get_nmr_inst_per_problem()
        eval_func_name = func_name + '_evaluateModel'
        obs_func_name = func_name + '_getObservation'

        param_listing = ''
        for p in self._evaluation_model.get_free_parameters():
            param_listing += self._get_param_listing_for_param(self._evaluation_model, p)

        func = ''
        func += self._evaluation_model.get_cl_dependency_headers()
        func += self._evaluation_model.get_cl_dependency_code()

        func += self.get_model_eval_function(eval_func_name)
        func += self.get_observation_return_function(obs_func_name)
        func += str(self._evaluation_model.get_objective_function(func_name, inst_per_problem, eval_func_name,
                                                                  obs_func_name, param_listing))
        return str(func)

    def get_objective_list_function(self, func_name="calculateObjectiveList"):
        inst_per_problem = self.get_nmr_inst_per_problem()
        eval_func_name = func_name + '_evaluateModel'
        obs_func_name = func_name + '_getObservation'

        param_listing = ''
        for p in self._evaluation_model.get_free_parameters():
            param_listing += self._get_param_listing_for_param(self._evaluation_model, p)

        func = ''
        func += self._evaluation_model.get_cl_dependency_headers()
        func += self._evaluation_model.get_cl_dependency_code()

        func += self.get_model_eval_function(eval_func_name)
        func += self.get_observation_return_function(obs_func_name)
        func += str(self._evaluation_model.get_objective_list_function(func_name, inst_per_problem, eval_func_name,
                                                                       obs_func_name, param_listing))
        return str(func)

    def finalize_optimization_results(self, results_dict):
        """This adds the final optimization maps to the results dictionary.

        Steps in finalizing the results dict:
            1) It first adds the maps for the dependent and fixed parameters
            2) Second it adds the extra maps defined in the models itself.
            3) Third it loops through the post_optimization_modifiers callback functions for the final updates.
            4) Adds additional maps defined in this model subclass

        For more documentation see the base method.

        Args:
            results_dict (dict): the dictionary with the results (the keys are the parameter names, the values are the
                1d parameter lists)

        """
        self._add_dependent_parameter_maps(results_dict)
        self._add_fixed_parameter_maps(results_dict)

        for model in self._model_functions_info.get_model_list():
            results_dict.update(model.get_extra_results_maps(results_dict))

        for name, routine in self._post_optimization_modifiers:
            results_dict[name] = routine(results_dict)

        self._add_finalizing_result_maps(results_dict)

        return results_dict

    def _add_fixed_parameter_maps(self, results_dict):
        """In place add complete maps for the fixed parameters."""
        param_lists = self._get_parameter_type_lists()
        fixed_params = param_lists['fixed']
        for (m, p) in fixed_params:
            if not self._model_functions_info.is_parameter_fixed_to_dependency(m, p):
                name = m.name + '.' + p.name
                value = self._parameter_values['{}.{}'.format(m.name, p.name)]

                if is_scalar(value):
                    results_dict.update({name: np.tile(np.array([value]), (self.get_nmr_problems(),))})
                else:
                    if self.problems_to_analyze is not None:
                        value = value[self.problems_to_analyze, ...]
                    results_dict.update({name: value})

    def _add_dependent_parameter_maps(self, results_dict):
        """In place add complete maps for the dependent parameters."""
        param_lists = self._get_parameter_type_lists()
        if len(param_lists['dependent']):
            func = ''
            func += self._get_fixed_parameters_listing(param_lists['fixed'])
            func += self._get_estimable_parameters_listing(param_lists['estimable'])
            func += self._get_dependent_parameters_listing(param_lists['dependent'])

            estimable_params = [m.name + '.' + p.name for m, p in param_lists['estimable']]
            estimated_parameters = [results_dict[k] for k in estimable_params]

            dependent_parameter_names = [(m.name + '_' + p.name, m.name + '.' + p.name)
                                         for m, p in param_lists['dependent']]

            cpd = CalculateDependentParameters(double_precision=self.double_precision)
            dependent_parameters = cpd.calculate(self, estimated_parameters, func, dependent_parameter_names)

            results_dict.update(dependent_parameters)

    def _get_parameter_transformations(self):
        dep_list = {}
        transform_dict = {}
        for m, p in self._model_functions_info.get_estimable_parameters_list():
            name = m.name + '.' + p.name
            dep_names = list(dep[0].name + '.' + dep[1].name for dep in p.parameter_transform.dependencies)
            dep_list.update({name: dep_names})
            transform_dict.update({name: p})

        dep_list = TopologicalSort(dep_list).get_flattened()

        dec_func_list = []
        enc_func_list = []
        for name in dep_list:
            parameter = transform_dict[name]
            ind = self._model_functions_info.get_parameter_estimable_index(name)
            transform = parameter.parameter_transform

            dependency_names = []
            for dep in transform.dependencies:
                dep_ind = self._model_functions_info.get_parameter_estimable_index(dep[0].name + '.' + dep[1].name)
                dependency_names.append('{0}[' + str(dep_ind) + ']')

            if all_elements_equal(self._lower_bounds[name]):
                lower_bound = str(get_single_value(self._lower_bounds[name]))
            else:
                lower_bound = 'data->var_data_lb_' + name.replace('.', '_')

            if all_elements_equal(self._upper_bounds[name]):
                upper_bound = str(get_single_value(self._upper_bounds[name]))
            else:
                upper_bound = 'data->var_data_ub_' + name.replace('.', '_')

            s = '{0}[' + str(ind) + '] = ' + transform.get_cl_decode().create_assignment(
                '{0}[' + str(ind) + ']', lower_bound, upper_bound, dependency_names) + ';'

            dec_func_list.append(s)

            s = '{0}[' + str(ind) + '] = ' + transform.get_cl_encode().create_assignment(
                '{0}[' + str(ind) + ']', lower_bound, upper_bound, dependency_names) + ';'

            enc_func_list.append(s)

        return tuple(reversed(enc_func_list)), dec_func_list

    def _add_finalizing_result_maps(self, results_dict):
        """Add some final results maps to the results dictionary.

        This called by the function finalize_optimization_results() as last call to add more maps.

        Args:
            results_dict (args): the results from model optmization. We are to modify this in-place.
        """

    def _transform_observations(self, observations):
        """Apply a transformation on the observations before fitting.

        This function is called by get_problems_var_data() just before the observations are handed over to the
        CL routine, and just after the the list has been (optionally) limited with self.problems_to_analyze.

        To implement any behaviour here, you can override this function and add behaviour that changes the observations.

        Args:
            observations (ndarray): the 2d matrix with the observations. This is the list of
                observations *after* the list has been (optionally) limited with self.problems_to_analyze.

        Returns:
            observations (ndarray): a 2d matrix of the same shape as the input. This should hold the transformed data.
        """
        return observations

    def _construct_model_expression(self, noise_func_name):
        """Construct the model signel expression. This is supposed to be used in get_model_eval_function.

        Args:
            noise_func_name (str): the name of the noise function.
        """
        func = ''
        if self._signal_noise_model:
            noise_params_func = ''
            for p in self._signal_noise_model.get_free_parameters():
                noise_params_func += ', ' + self._signal_noise_model.name + '_' + p.name

            func += noise_func_name + '((' + self._build_model_from_tree(self._model_tree, 0) + ')' + \
                    noise_params_func + ');'
        else:
            func += '(' + self._build_model_from_tree(self._model_tree, 0) + ');'
        return func

    def _build_model_from_tree(self, node, depth):
        if not node.children:
            return self._model_to_string(node.data)
        else:
            subfuncs = []
            for child in node.children:
                if child.children:
                    subfuncs.append(self._build_model_from_tree(child, depth+1))
                else:
                    subfuncs.append(self._model_to_string(child.data))

            operator = node.data
            func = (' ' + operator + ' ').join(subfuncs)

        if func[0] == '(':
            return '(' + func + ')'
        return '(' + "\n" + ("\t" * int((depth/2)+5)) + func + "\n" + ("\t" * int((depth/2)+4)) + ')'

    def _model_to_string(self, model):
        """Convert a model to CL string."""
        param_list = []
        for param in model.parameter_list:
            if isinstance(param, ProtocolParameter):
                param_list.append(param.name)
            elif isinstance(param, ModelDataParameter):
                value = self._parameter_values['{}.{}'.format(model.name, param.name)]
                if all_elements_equal(value):
                    param_list.append(str(get_single_value(value)))
                else:
                    param_list.append('data->model_data_' + param.name)
            elif isinstance(param, StaticMapParameter):
                static_map_value = self._get_static_map_value(model, param)
                if all_elements_equal(static_map_value):
                    param_list.append(str(get_single_value(static_map_value)))
                else:
                    if len(static_map_value.shape) > 1 \
                            and static_map_value.shape[1] == self._problem_data.observations.shape[1]:
                        param_list.append('data->var_data_' + model.name + '_' + param.name + '[observation_index]')
                    else:
                        param_list.append('data->var_data_' + model.name + '_' + param.name)
            elif isinstance(param, CurrentObservationParam):
                param_list.append('data->var_data_observations[observation_index]')
            else:
                param_list.append(model.name + '_' + param.name)

        return model.cl_function_name + '(' + ', '.join(param_list) + ')'

    def _get_model_functions_cl_code(self, noise_func_name):
        """Get the model functions CL. This is used in get_model_eval_function()."""
        cl_code = ''
        for compartment_model in self._model_tree.get_compartment_models():
            cl_code += compartment_model.get_cl_header() + "\n"
            cl_code += compartment_model.get_cl_code() + "\n"

        if self._signal_noise_model:
            cl_code += self._signal_noise_model.get_signal_function(noise_func_name)
        return cl_code

    def _get_parameters_listing(self, exclude_list=()):
        """Get the CL code for the parameter listing, this goes on top of the evaluate function.

        Args:
            exclude_list: an optional list containing parameters to exclude from the listing.
             This should contain full parameter names like: <model_name>_<param_name>

        Returns:
            An CL string that contains all the parameters as primitive data types.
        """
        func = ''
        param_lists = self._get_parameter_type_lists()
        func += self._get_protocol_parameters_listing(param_lists['protocol'], exclude_list=exclude_list)
        func += self._get_fixed_parameters_listing(param_lists['fixed'], exclude_list=exclude_list)
        func += self._get_estimable_parameters_listing(param_lists['estimable'], exclude_list=exclude_list)
        func += self._get_dependent_parameters_listing(param_lists['dependent'], exclude_list=exclude_list)
        return str(func)

    def _get_estimable_parameters_listing(self, param_list=None, exclude_list=()):
        """Get the parameter listing for the free parameters.

        For performance reasons, the parameter list should already be given.
            If not given it is calculated using:
                self._get_parameter_type_lists()['estimable']

        Args:
            param_list: the list with the estimable parameters
            exclude_list: a list of parameters to exclude from this listing
        """
        if param_list is None:
            param_list = self._get_parameter_type_lists()['estimable']

        func = ''
        estimable_param_counter = 0
        for m, p in param_list:
            name = m.name + '_' + p.name
            if name not in exclude_list:
                data_type = p.data_type.cl_type
                assignment = 'x[' + str(estimable_param_counter) + ']'
                func += "\t"*4 + data_type + ' ' + name + ' = ' + assignment + ';' + "\n"
                estimable_param_counter += 1
        return func

    def _get_protocol_parameters_listing(self, param_list=None, exclude_list=()):
        """Get the parameter listing for the protocol parameters.

        For performance reasons, the parameter list should already be given.
            If not given it is calculated using:
                self._get_parameter_type_lists()['protocol']

        Args:
            param_list: the list with the protocol parameters
            exclude_list: a list of parameters to exclude from this listing
        """
        protocol_info = self._problem_data.protocol
        if param_list is None:
            param_list = self._get_parameter_type_lists()['protocol']

        const_params_seen = []
        func = ''
        for m, p in param_list:
            if (m.name + '_' + p.name) not in exclude_list:
                data_type = p.data_type.cl_type
                if p.name not in const_params_seen:
                    if all_elements_equal(protocol_info[p.name]):
                        if p.data_type.is_vector_type:
                            vector_length = p.data_type.vector_length
                            values = [str(val) for val in protocol_info[p.name][0]]
                            if len(values) < vector_length:
                                values.append(str(0))
                            assignment = '(' + data_type + ')(' + ', '.join(values) + ')'
                        else:
                            assignment = str(float(protocol_info[p.name][0]))
                    else:
                        if p.data_type.is_pointer_type:
                            # this requires generic address spaces available in OpenCL >= 2.0.
                            assignment = '&data->protocol_data_' + p.name + '[observation_index]'
                        else:
                            assignment = 'data->protocol_data_' + p.name + '[observation_index]'
                    func += "\t"*4 + data_type + ' ' + p.name + ' = ' + assignment + ';' + "\n"
                    const_params_seen.append(p.name)
        return func

    def _get_fixed_parameters_listing(self, param_list=None, exclude_list=()):
        """Get the parameter listing for the fixed parameters.

        For performance reasons, the fixed parameter list should already be given.
            If not given it is calculated using:
                self._get_parameter_type_lists()['fixed']

        Args:
            dependent_param_list: the list list of fixed params
            exclude_list: a list of parameters to exclude from this listing
        """
        if param_list is None:
            param_list = self._get_parameter_type_lists()['fixed']

        func = ''
        for m, p in param_list:
            name = m.name + '_' + p.name
            if name not in exclude_list:
                data_type = p.data_type.raw_data_type
                value = self._parameter_values['{}.{}'.format(m.name, p.name)]

                if all_elements_equal(value):
                    assignment = '(' + data_type + ')' + str(float(get_single_value(value)))
                else:
                    assignment = '(' + data_type + ') data->var_data_' + m.name + '_' + p.name
                func += "\t"*4 + data_type + ' ' + name + ' = ' + assignment + ';' + "\n"
        return func

    def _get_dependent_parameters_listing(self, dependent_param_list=None, exclude_list=()):
        """Get the parameter listing for the dependent parameters.

        For performance reasons, the dependent parameter list should already be given.
            If not given it is calculated using:
                self._get_parameter_type_lists()['dependent']

        Args:
            dependent_param_list: the list list of dependent params
            exclude_list: a list of parameters to exclude from this listing, note that this will only exclude the
                definition of the parameter, not the dependency code.
        """
        if dependent_param_list is None:
            dependent_param_list = self._get_parameter_type_lists()['dependent']
        func = ''
        for m, p in dependent_param_list:
            pd = self._dependency_store.get_dependency(m.name + '.' + p.name)
            if pd.pre_transform_code:
                func += "\t"*4 + self._convert_parameters_dot_to_bar(pd.pre_transform_code)

            assignment = self._convert_parameters_dot_to_bar(pd.assignment_code)
            name = m.name + '_' + p.name
            data_type = p.data_type.raw_data_type

            if self._model_functions_info.is_parameter_fixed_to_dependency(m, p):
                if (m.name + '_' + p.name) not in exclude_list:
                    func += "\t"*4 + data_type + ' ' + name + ' = ' + assignment + ';' + "\n"
            else:
                func += "\t"*4 + name + ' = ' + assignment + ';' + "\n"
        return func

    def _get_fixed_parameters_as_var_data(self):
        var_data_dict = {}
        for m, p in self._model_functions_info.get_free_parameters_list():
            value = self._parameter_values['{}.{}'.format(m.name, p.name)]

            if self._model_functions_info.is_fixed_to_value('{}.{}'.format(m.name, p.name)) \
                and not all_elements_equal(value) \
                and not self._model_functions_info.is_parameter_fixed_to_dependency(m, p):

                if self.problems_to_analyze is not None:
                    value = value[self.problems_to_analyze, ...]

                var_data_dict.update({m.name + '_' + p.name: SimpleDataAdapter(value, p.data_type,
                                                                               self._get_mot_float_type())})
        return var_data_dict

    def _get_static_parameters_as_var_data(self):
        static_data_dict = {}

        for m, p in self._model_functions_info.get_static_parameters_list():
            static_map_value = self._get_static_map_value(m, p)

            if not all_elements_equal(static_map_value):
                data_adapter = SimpleDataAdapter(static_map_value, p.data_type, self._get_mot_float_type())
                static_data_dict.update({m.name + '_' + p.name: data_adapter})

        return static_data_dict

    def _get_bounds_as_var_data(self):
        bounds_dict = {}

        for m, p in self._model_functions_info.get_free_parameters_list():
            lower_bound = self._lower_bounds['{}.{}'.format(m.name, p.name)]
            upper_bound = self._upper_bounds['{}.{}'.format(m.name, p.name)]

            if not all_elements_equal(lower_bound):
                data_adapter = SimpleDataAdapter(lower_bound, p.data_type, self._get_mot_float_type())
                bounds_dict.update({'lb_' + m.name + '_' + p.name: data_adapter})

            if not all_elements_equal(upper_bound):
                data_adapter = SimpleDataAdapter(upper_bound, p.data_type, self._get_mot_float_type())
                bounds_dict.update({'ub_' + m.name + '_' + p.name: data_adapter})

        return bounds_dict

    def _get_static_map_value(self, model, parameter):
        """Get the map value for the given parameter of the given model.

        This first checks if the parameter is defined in the static maps data in the problem data. If not, we try
        to get it from the value stored in the parameter itself. If that fails as well we raise an error.

        Also, this only returns the problems for which problems_to_analyze is set.

        Args:
            model (ModelFunction): the model function
            parameter (CLParameter): the parameter for which we want to get the value

        Returns:
            ndarray or number: the value for the given parameter.
        """
        data = None
        value = self._parameter_values.get('{}.{}'.format(model.name, parameter.name), None)
        if parameter.name in self._problem_data.static_maps:
            data = self._problem_data.static_maps[parameter.name]
        elif value is not None:
            data = value

        if data is None:
            raise ValueError('No suitable data could be found for the static parameter {}.'.format(parameter.name))

        if is_scalar(data):
            return data

        if self.problems_to_analyze is not None:
            return data[self.problems_to_analyze, ...]
        return data

    def _is_non_model_tree_model(self, model):
        return model is self._evaluation_model or (self._signal_noise_model is not None and
                                                   model is self._signal_noise_model)

    def _get_param_listing_for_param(self, m, p):
        """Get the param listing for one specific parameter. This can be used for example for the noise model params.

        Please note, that on the moment this function does not support the complete dependency graph for the dependent
        parameters.
        """
        data_type = p.data_type.raw_data_type
        name = m.name + '_' + p.name
        assignment = ''

        if isinstance(p, ProtocolParameter):
            assignment = 'data->protocol_data_' + p.name + '[observation_index]'
        elif isinstance(p, FreeParameter):
            value = self._parameter_values['{}.{}'.format(m.name, p.name)]

            if self._model_functions_info.is_fixed_to_value('{}.{}'.format(m.name, p.name)) \
                and not self._model_functions_info.parameter_has_dependency(m, p):
                if all_elements_equal(value):
                    assignment = '(' + data_type + ')' + str(float(get_single_value(value)))
                else:
                    assignment = '(' + data_type + ') data->var_data_' + m.name + '_' + p.name
            elif not self._model_functions_info.parameter_has_dependency(m, p) \
                or (self._model_functions_info.parameter_has_dependency(m, p)
                    and not self._model_functions_info.is_parameter_fixed_to_dependency(m, p)):
                ind = self._model_functions_info.get_parameter_estimable_index(m.name + '.' + p.name)
                assignment += 'x[' + str(ind) + ']'
            if self._model_functions_info.parameter_has_dependency(m, p):
                return self._get_dependent_parameters_listing(((m, p),))

        return data_type + ' ' + name + ' = ' + assignment + ';' + "\n"

    def _get_parameter_type_lists(self):
        """Returns a dictionary with the parameters sorted in the types protocol, fixed, estimable and dependent.

        Parameters may occur in different lists (estimable and dependent for example).
        """
        protocol_parameters = []
        fixed_parameters = []
        estimable_parameters = []
        depended_parameters = []

        for m, p in self._model_functions_info.get_model_parameter_list():
            if isinstance(p, ProtocolParameter):
                protocol_parameters.append((m, p))
            elif isinstance(p, FreeParameter):
                if self._model_functions_info.is_fixed_to_value('{}.{}'.format(m.name, p.name)) \
                    and not self._model_functions_info.parameter_has_dependency(m, p):
                    fixed_parameters.append((m, p))
                elif not self._model_functions_info.parameter_has_dependency(m, p) \
                    or (self._model_functions_info.parameter_has_dependency(m, p)
                        and not self._model_functions_info.is_parameter_fixed_to_dependency(m, p)):
                    estimable_parameters.append((m, p))

                if self._model_functions_info.parameter_has_dependency(m, p):
                    ind = self._dependency_store.get_index(m.name + '.' + p.name)
                    depended_parameters.insert(ind, (m, p))

        return {'protocol': protocol_parameters, 'fixed': fixed_parameters,
                'estimable': estimable_parameters, 'dependent': depended_parameters}

    def _convert_parameters_dot_to_bar(self, string):
        """Convert a string containing parameters with . to parameter names with _"""
        for dname, bname in self._parameters_dot_to_bar.items():
            string = string.replace(dname, bname)
        return string

    def _init_fixed_duplicates_dependencies(self):
        """Find duplicate fixed parameters, and make dependencies of them. This saves data transfer in CL."""
        var_data_dict = {}
        for m, p in self._model_functions_info.get_free_parameters_list():
            value = self._parameter_values['{}.{}'.format(m.name, p.name)]

            if self._model_functions_info.is_fixed_to_value('{}.{}'.format(m.name, p.name)) \
                    and not is_scalar(value) \
                    and not self._model_functions_info.is_parameter_fixed_to_dependency(m, p):

                duplicate_found = False
                duplicate_key = None

                for key, data in var_data_dict.items():
                    if np.array_equal(data, value):
                        duplicate_found = True
                        duplicate_key = key
                        break

                if duplicate_found:
                    self.add_parameter_dependency(m.name + '.' + p.name, SimpleAssignment(duplicate_key))
                else:
                    var_data_dict.update({m.name + '.' + p.name: value})

    def _get_variable_data(self):
        """See super class OptimizeModelInterface for details

        When overriding this function, please note that it should adhere to the attribute problems_to_analyze.
        """
        var_data_dict = {}

        observations = self._problem_data.observations
        if observations is not None:
            if self.problems_to_analyze is not None:
                observations = observations[self.problems_to_analyze, ...]

            observations = self._transform_observations(observations)

            data_adapter = SimpleDataAdapter(observations, CLDataType.from_string('mot_float_type*'),
                                             self._get_mot_float_type())
            var_data_dict.update({'observations': data_adapter})

        var_data_dict.update(self._get_fixed_parameters_as_var_data())
        var_data_dict.update(self._get_static_parameters_as_var_data())
        var_data_dict.update(self._get_bounds_as_var_data())

        return var_data_dict

    def _get_protocol_data(self):
        protocol_info = self._problem_data.protocol
        return_data = {}
        for m, p in self._model_functions_info.get_model_parameter_list():
            if isinstance(p, ProtocolParameter):
                if p.name in protocol_info:
                    if not all_elements_equal(protocol_info[p.name]):
                        const_d = {p.name: SimpleDataAdapter(protocol_info[p.name],
                                                             p.data_type, self._get_mot_float_type())}
                        return_data.update(const_d)
                else:
                    exception = 'Protocol parameter "{}" could not be resolved'.format(m.name + '.' + p.name)
                    raise ParameterResolutionException(exception)
        return return_data

    def _get_static_data(self):
        model_data_dict = {}
        for m, p in self._model_functions_info.get_model_parameter_list():
            if isinstance(p, ModelDataParameter):
                value = self._parameter_values['{}.{}'.format(m.name, p.name)]
                if not all_elements_equal(value):
                    model_data_dict.update({p.name: SimpleDataAdapter(value, p.data_type, self._get_mot_float_type())})
        return model_data_dict

    def _get_all_kernel_source_items(self, device):
        """Get the CL strings for the kernel source items for most common CL kernels in this library."""
        import pyopencl as cl

        max_constant_buffer_size = device.get_info(cl.device_info.MAX_CONSTANT_BUFFER_SIZE)
        max_constant_args = device.get_info(cl.device_info.MAX_CONSTANT_ARGS)

        def _check_array_fits_constant_buffer(array, dtype):
            """Check if the given array when casted to the given type can be fit into the given max_size

            Args:
                array (ndarray): the array we want to fit
                dtype (np data type): the numpy data type we want to use

            Returns:
                boolean: if it fits in the constant memory buffer or not
            """
            return np.product(array.shape) * np.dtype(dtype).itemsize < max_constant_buffer_size

        constant_args_counter = 0

        kernel_param_names = []
        data_struct_init = []
        data_struct_names = []

        for key, data_adapter in self._get_variable_data().items():
            clmemtype = 'global'

            cl_data = data_adapter.get_opencl_data()

            if _check_array_fits_constant_buffer(cl_data, data_adapter.get_opencl_numpy_type()):
                if constant_args_counter < max_constant_args:
                    clmemtype = 'constant'
                    constant_args_counter += 1

            param_name = 'var_data_' + str(key)
            data_type = data_adapter.get_data_type().raw_data_type

            if data_adapter.get_data_type().is_vector_type:
                data_type += data_adapter.get_data_type().vector_length

            kernel_param_names.append(clmemtype + ' ' + data_type + '* ' + param_name)

            mult = cl_data.shape[1] if len(cl_data.shape) > 1 else 1
            if len(cl_data.shape) == 1 or cl_data.shape[1] == 1:
                data_struct_names.append(data_type + ' ' + param_name)
                data_struct_init.append(param_name + '[gid * ' + str(mult) + ']')
            else:
                data_struct_names.append(clmemtype + ' ' + data_type + '* ' + param_name)
                data_struct_init.append(param_name + ' + gid * ' + str(mult))

        for key, data_adapter in self._get_protocol_data().items():
            clmemtype = 'global'

            cl_data = data_adapter.get_opencl_data()

            if _check_array_fits_constant_buffer(cl_data, data_adapter.get_opencl_numpy_type()):
                if constant_args_counter < max_constant_args:
                    clmemtype = 'constant'
                    constant_args_counter += 1

            param_name = 'protocol_data_' + str(key)
            data_type = data_adapter.get_data_type().raw_data_type

            if data_adapter.get_data_type().is_vector_type:
                data_type += str(data_adapter.get_data_type().vector_length)

            kernel_param_names.append(clmemtype + ' ' + data_type + '* ' + param_name)
            data_struct_init.append(param_name)
            data_struct_names.append(clmemtype + ' ' + data_type + '* ' + param_name)

        for key, data_adapter in self._get_static_data().items():
            clmemtype = 'global'
            param_name = 'model_data_' + str(key)
            data_type = data_adapter.get_data_type().raw_data_type

            if data_adapter.get_data_type().is_vector_type:
                data_type += data_adapter.get_data_type().vector_length

            data_struct_init.append(param_name)

            if isinstance(data_adapter.get_opencl_data(), np.ndarray):
                kernel_param_names.append(clmemtype + ' ' + data_type + '* ' + param_name)
                data_struct_names.append(clmemtype + ' ' + data_type + '* ' + param_name)
            else:
                kernel_param_names.append(data_type + ' ' + param_name)
                data_struct_names.append(data_type + ' ' + param_name)

        data_struct = '''
            typedef struct{
                ''' + ('' if data_struct_names else 'constant void* place_holder;') + '''
                ''' + " ".join((name + ";\n" for name in data_struct_names)) + '''
            } ''' + self.get_kernel_data_struct_type() + ''';
        '''

        return {'kernel_param_names': kernel_param_names,
                'data_struct_names': data_struct_names,
                'data_struct_init': data_struct_init,
                'data_struct': data_struct}

    def _get_pre_model_expression_eval_code(self):
        """The code called in the evaluation function.

        This is called after the parameters are initialized and before the model signal expression. It can call
        functions defined in _get_pre_model_expression_eval_function()

        Returns:
            str: cl code containing evaluation changes,
        """
        return ''

    def _get_pre_model_expression_eval_function(self):
        """Function used in the model evaluation generation function.

        The idea is that some implementing models may need to change some of the protocol or fixed parameters
        before they are handed over to the signal expression function. This function is called by the
        get_model_eval_function function during model evaluation function construction.

        Returns:
            str: cl function to be used in conjunction with the output of the function
                _get_pre_model_expression_eval_model()
        """

    def _set_default_dependencies(self):
        """Initialize the default dependencies.

        By default this only adds dependencies for the fixed data that is used in multiple parameters.
        """
        self._init_fixed_duplicates_dependencies()

    def _get_mot_float_type(self):
        """Get the data type for the mot_float_type"""
        if self.double_precision:
            return CLDataType.from_string('double')
        return CLDataType.from_string('float')


class SampleModelBuilder(OptimizeModelBuilder, SampleModelInterface):

    def __init__(self, model_name, model_tree, evaluation_model, signal_noise_model=None, problem_data=None):
        super(SampleModelBuilder, self).__init__(model_name, model_tree, evaluation_model, signal_noise_model,
                                                 problem_data)

    def get_log_prior_function(self, func_name='getLogPrior'):
        prior = 'mot_float_type ' + func_name + '(const void* data_void, const mot_float_type* const x){' + "\n"
        prior += "\t" + self.get_kernel_data_struct_type() + \
                    '* data = (' + self.get_kernel_data_struct_type() + '*)data_void;\n'
        prior += '\tmot_float_type prior = 1.0;\n'

        for i, (m, p) in enumerate(self._model_functions_info.get_estimable_parameters_list()):
            name = m.name + '.' + p.name

            if all_elements_equal(self._lower_bounds[name]):
                lower_bound = str(get_single_value(self._lower_bounds[name]))
            else:
                lower_bound = 'data->var_data_lb_' + name.replace('.', '_')

            if all_elements_equal(self._upper_bounds[name]):
                upper_bound = str(get_single_value(self._upper_bounds[name]))
            else:
                upper_bound = 'data->var_data_ub_' + name.replace('.', '_')

            assignment_constructor = p.sampling_prior.get_cl_assignment()
            assignment = assignment_constructor.create_assignment('x[' + str(i) + ']', lower_bound, upper_bound)

            prior += '\tprior *= {};\n'.format(assignment)

        prior += '\n\treturn log(prior);\n}'
        return prior

    def get_proposal_state(self):
        return_list = []
        for m, p in self._model_functions_info.get_estimable_parameters_list():
            for param in p.sampling_proposal.get_parameters():
                if param.adaptable:
                    return_list.append(param.default_value)
        return return_list

    def is_proposal_symmetric(self):
        return all(p.sampling_proposal.is_symmetric for m, p in
                   self._model_functions_info.get_estimable_parameters_list())

    def get_proposal_logpdf(self, func_name='getProposalLogPDF'):
        return_str = ''
        for _, p in self._model_functions_info.get_estimable_parameters_list():
            return_str += p.sampling_proposal.get_proposal_logpdf_function()

        return_str += "\n" + 'mot_float_type ' + func_name + \
            '(const int i, const mot_float_type proposal, const mot_float_type current, ' \
            ' mot_float_type* const proposal_state){' + "\n\t"

        return_str += "\n\t" + 'switch(i){' + "\n\t\t"

        adaptable_parameter_count = 0
        for i, (m, p) in enumerate(self._model_functions_info.get_estimable_parameters_list()):
            return_str += 'case ' + str(i) + ':' + "\n\t\t\t"

            param_proposal = p.sampling_proposal
            logpdf_call = 'return ' + param_proposal.get_proposal_logpdf_function_name() + '(proposal, current'

            for param in param_proposal.get_parameters():
                if param.adaptable:
                    logpdf_call += ', proposal_state[' + str(adaptable_parameter_count) + ']'
                    adaptable_parameter_count += 1
                else:
                    logpdf_call += ', ' + str(param.default_value)

            logpdf_call += ');'
            return_str += logpdf_call + "\n"

        return_str += "\n\t\t" + '}' + "\n" + 'return 0;' + "\n"
        return_str += '}'
        return return_str

    def get_proposal_function(self, func_name='getProposal'):
        return_str = ''
        for _, p in self._model_functions_info.get_estimable_parameters_list():
            return_str += p.sampling_proposal.get_proposal_function()

        return_str += "\n" + 'mot_float_type ' + func_name + \
            '(const int i, const mot_float_type current, void* rng_data, ' \
            ' mot_float_type* const proposal_state){'

        return_str += "\n\t" + 'switch(i){' + "\n\t\t"

        adaptable_parameter_count = 0
        for i, (m, p) in enumerate(self._model_functions_info.get_estimable_parameters_list()):
            return_str += 'case ' + str(i) + ':' + "\n\t\t\t"

            param_proposal = p.sampling_proposal
            proposal_call = 'return ' + param_proposal.get_proposal_function_name() + '(current, rng_data'

            for param in param_proposal.get_parameters():
                if param.adaptable:
                    proposal_call += ', proposal_state[' + str(adaptable_parameter_count) + ']'
                    adaptable_parameter_count += 1
                else:
                    proposal_call += ', ' + str(param.default_value)

            proposal_call += ');'
            return_str += proposal_call + "\n"

        return_str += "\n\t\t" + '}' + "\n" + 'return 0;' + "\n"
        return_str += '}'
        return return_str

    def get_proposal_state_update_function(self, func_name='updateProposalParameters'):
        return_str = ''
        for _, p in self._model_functions_info.get_estimable_parameters_list():
            for param in p.sampling_proposal.get_parameters():
                if param.adaptable:
                    return_str += param.get_parameter_update_function()

        return_str += 'void ' + func_name + '(uint* const ac_between_proposal_updates, ' + \
            'const uint proposal_update_intervals, mot_float_type* const proposal_state){' + "\n"

        adaptable_parameter_count = 0
        for i, (m, p) in enumerate(self._model_functions_info.get_estimable_parameters_list()):
            param_proposal = p.sampling_proposal

            for param in param_proposal.get_parameters():
                if param.adaptable:
                    return_str += "\t" * 3
                    return_str += 'proposal_state[' + str(adaptable_parameter_count) + '] = '
                    return_str += param.get_parameter_update_function_name() + '(' +\
                        'proposal_state[' + str(adaptable_parameter_count) + '], ' + \
                        'ac_between_proposal_updates[' + str(i) + '], proposal_update_intervals);' + "\n"
                    adaptable_parameter_count += 1

        return_str += '}'
        return return_str

    def get_log_likelihood_function(self, func_name='getLogLikelihood', evaluation_model=None, full_likelihood=True):
        evaluation_model = evaluation_model or self._evaluation_model

        inst_per_problem = self.get_nmr_inst_per_problem()
        eval_func_name = func_name + '_evaluateModel'
        obs_func_name = func_name + '_getObservation'

        param_listing = ''
        for p in evaluation_model.get_free_parameters():
            param_listing += self._get_param_listing_for_param(evaluation_model, p)

        func = ''
        func += evaluation_model.get_cl_dependency_headers()
        func += evaluation_model.get_cl_dependency_code()

        func += self.get_model_eval_function(eval_func_name)
        func += self.get_observation_return_function(obs_func_name)
        func += evaluation_model.get_log_likelihood_function(func_name, inst_per_problem, eval_func_name,
                                                             obs_func_name, param_listing,
                                                             full_likelihood=full_likelihood)
        return func

    def samples_to_statistics(self, samples_dict):
        results = {}
        for key, value in samples_dict.items():
            _, param = self._model_functions_info.get_model_parameter_by_name(key)
            stat_mod = param.sampling_statistics
            results[key] = stat_mod.get_mean(value)
            results[key + '.std'] = stat_mod.get_std(value)
        return results


class _DependencyStore(object):

    def __init__(self):
        self.names_in_order = []
        self.dependencies = {}

    def set_dependency(self, param_name, dependency):
        if param_name not in self.names_in_order:
            self.names_in_order.append(param_name)
        self.dependencies.update({param_name: dependency})

    def get_dependency(self, param_name):
        return self.dependencies[param_name]

    def has_dependency(self, param_name):
        return param_name in self.dependencies

    def has_dependencies(self):
        return self.names_in_order

    def get_index(self, param_name):
        return self.names_in_order.index(param_name)


class _ModelFunctionsInformation(object):

    def __init__(self, dependency_store, model_tree, evaluation_model, signal_noise_model=None):
        """Contains centralized information about the model functions in the model builder parent.

        Args:
            dependency_store (_DependencyStore): the shared dependency store containing information about
                the parameters that depend on each other
            model_tree (mot.model_building.trees.CompartmentModelTree): the model tree object
            evaluation_model (mot.model_building.evaluation_models.EvaluationModel): the evaluation model to
                use for the resulting complete model
            signal_noise_model (mot.model_building.signal_noise_models.SignalNoiseModel): the signal
                noise model to use to add noise to the model prediction
        """
        self._dependency_store = dependency_store
        self._model_tree = model_tree
        self._evaluation_model = evaluation_model
        self._signal_noise_model = signal_noise_model

        self._model_list = self._get_model_list()
        self._model_parameter_list = self._get_model_parameter_list()

        self._check_for_double_model_names()

        self._fixed_parameters = {'{}.{}'.format(m.name, p.name): p.fixed for m, p in self.get_free_parameters_list()}

    def set_fixed_to_value(self, parameter_name, fix_state):
        """Set the given parameter fixed.

        This only works with free parameters.

        Args:
            parameter_name (str): the name of the parameter to fix or unfix
            fix_state (boolean): if the parameter is fixed or not
        """
        self._fixed_parameters[parameter_name] = fix_state

    def is_fixed_to_value(self, parameter_name):
        """Check if the given (free) parameter is fixed to a value.

        Args:
            parameter_name (str): the name of the parameter to fix or unfix

        Returns:
            boolean: if the parameter is fixed to a value or not
        """
        return self._fixed_parameters[parameter_name]

    def get_model_list(self):
        """Get the list of all the applicable model functions

        Returns:
            list of mot.model_building.cl_functions.base.ModelFunction: the list of model functions.
        """
        return self._model_list

    def get_model_parameter_list(self):
        """Get a list of all model, parameter tuples.

        Returns:
            list of tuple: the list of tuples containing (model, parameters)
        """
        return self._model_parameter_list

    def get_free_parameters_list(self):
        """Gets the free parameters as (model, parameter) tuples from the model listing.
        This does not incorporate checking for fixed parameters.

        Returns:
            list of tuple: the list of tuples containing (model, parameters)
        """
        return list((m, p) for m, p in self._model_parameter_list if isinstance(p, FreeParameter))

    def get_static_parameters_list(self):
        """Gets the static parameters (as model, parameter tuples) from the model listing."""
        return list((m, p) for m, p in self._model_parameter_list if isinstance(p, StaticMapParameter))

    def get_model_parameter_by_name(self, parameter_name):
        """Get the parameter object of the given full parameter name in dot format.

        Args:
            parameter_name (string): the parameter name in dot format: <model>.<param>

        Returns:
            tuple: containing the (model, parameter) pair for the given parameter name
        """
        for m, p in self._model_parameter_list:
            if '{}.{}'.format(m.name, p.name) == parameter_name:
                return m, p
        raise ValueError('The parameter with the name "{}" could not be found in this model.'.format(parameter_name))

    def get_non_model_tree_param_listing(self):
        """Get the model, parameter tuples for all parameters not in the model tree.

        Basically this returns the parameters of the evaluation and signal noise model.

        Returns:
            tuple: the (model, parameter) tuple for all non model tree parameters
        """
        listing = []
        for p in self._evaluation_model.parameter_list:
            listing.append((self._evaluation_model, p))

        if self._signal_noise_model:
            for p in self._signal_noise_model.parameter_list:
                listing.append((self._signal_noise_model, p))

        return listing

    def parameter_has_dependency(self, model, param):
        """Check if the given model and parameter name combo has a dependency.

        Args:
            model (mot.model_building.cl_functions.base.ModelFunction): the model function
            param (mot.model_building.cl_functions.parameters.CLFunctionParameter): the parameter

        Returns:
            boolean: if the given parameter has a dependency
        """
        return self._dependency_store.has_dependency(model.name + '.' + param.name)

    def is_parameter_fixed_to_dependency(self, model, param):
        """Check if the given model and parameter name combo has a dependency.

        Args:
            model (mot.model_building.cl_functions.base.ModelFunction): the model function
            param (mot.model_building.cl_functions.parameters.CLFunctionParameter): the parameter

        Returns:
            boolean: if the given parameter is fixed to a dependency. Returns False if either this parameter
                has no dependency or if it is not fixed to it.
        """
        return self.parameter_has_dependency(model, param) \
            and self._dependency_store.get_dependency(model.name + '.' + param.name).fixed

    def is_parameter_estimable(self, model, param):
        """Check if the given model parameter is estimable.

        A parameter is estimable if it is free, not fixed to dependencies and not fixed to any static value.

        Args:
            model (mot.model_building.cl_functions.base.ModelFunction): the model function
            param (mot.model_building.cl_functions.parameters.CLFunctionParameter): the parameter

        Returns:
            boolean: true if the parameter is estimable, false otherwise
        """
        return isinstance(param, FreeParameter) and \
            not self._fixed_parameters.get('{}.{}'.format(model.name, param.name), False) and \
            not self.is_parameter_fixed_to_dependency(model, param)

    def get_estimable_parameters_list(self):
        """Gets a list (as model, parameter tuples) of all parameters that are estimable.

        Returns:
            list of tuple: the list of estimable parameters
        """
        return [(m, p) for m, p in self._model_parameter_list if self.is_parameter_estimable(m, p)]

    def _get_model_parameter_list(self):
        """Get a list of all model, parameter tuples.

        Returns:
            list of tuple: the list of tuples containing (model, parameters)
        """
        return list((m, p) for m in self._model_list for p in m.parameter_list)

    def get_parameter_estimable_index(self, parameter_name):
        """Get the index of this parameter in the parameters list

        This returns the position of this parameter in the 'x', parameter vector in the CL kernels.

        Args:
            parameter_name: the parameter name in dot format. <model>.<param>

        Returns:
            int: the index of the requested parameter in the list of optimized parameters
        """
        estimable_param_counter = 0
        for m in self.get_model_list():
            for p in m.parameter_list:
                if '{}.{}'.format(m.name, p.name) == parameter_name:
                    if not self.is_parameter_estimable(m, p):
                        raise ValueError('The requested parameter "{}" is '
                                         'not an estimable parameter.'.format(parameter_name))
                    return estimable_param_counter

                if self.is_parameter_estimable(m, p):
                    estimable_param_counter += 1

        raise ValueError('The given parameter "{}" could not be found in this model'.format(parameter_name))

    def has_parameter(self, model_param_name):
        """Check to see if the given parameter is defined in this model.

        Args:
            model_param_name (string): A model.param name like 'Ball.d'

        Returns:
            boolean: true if the parameter is defined in this model, false otherwise.
        """
        for m, p in self.get_model_parameter_list():
            if '{}.{}'.format(m.name, p.name) == model_param_name:
                return True
        return False

    def _get_model_list(self):
        """Get the list of all the applicable model functions"""
        models = list(self._model_tree.get_compartment_models())
        models.append(self._evaluation_model)
        if self._signal_noise_model:
            models.append(self._signal_noise_model)
        return models

    def _check_for_double_model_names(self):
        models = self._model_list
        model_names = []
        for m in models:
            if m.name in model_names:
                raise DoubleModelNameException("Double model name detected in the model tree.", m.name)
            model_names.append(m.name)


class ParameterNameException(Exception):
    """Thrown when the a parameter of an given name could not be found."""
    pass


class ParameterResolutionException(Exception):
    """Thrown when a fixed parameter could not be resolved."""
    pass


class DoubleModelNameException(Exception):
    """Thrown when there are two models with the same name."""
    pass




__author__ = 'Robbert Harms'
__date__ = "2014-06-20"
__license__ = "LGPL v3"
__maintainer__ = "Robbert Harms"
__email__ = "robbert.harms@maastrichtuniversity.nl"


class AbstractTransformation(object):

    def __init__(self, dependencies=()):
        """The transformations define the encode and decode operations needed to build a codec.

        These objects define the basic transformation from and to model and optimization space.

        The state of the parameters is extrinsic. Calling the encode and decode functions needs a reference to a
        parameter, these should be handled by the client code.

        The transformation function may depend on other parameters. (Client code should handle the correct order
        of handling the transformations given all the dependencies.)

        Args:
            dependencies (list of (CLFunction, CLFunctionParameter) pairs): A list of (models, parameters)
                which are necessary for the operation of the transformation.
        """
        super(AbstractTransformation, self).__init__()
        self._dependencies = dependencies

    def get_cl_encode(self):
        """Get the CL encode assignment constructor

        Returns
            AssignmentConstructor: The cl code assignment constructor for encoding the parameter.
        """
        pass

    def get_cl_decode(self):
        """Get the CL decode assignment constructor

        Returns:
            AssignmentConstructor: The cl code assignment constructor for decoding the parameter.
        """
        pass

    @property
    def dependencies(self):
        """Get a list of (CLFunction, CLFunctionParameter) pairs where this transformation depends on

        Returns:
            list of tuple: A list of (CLFunction, CLFunctionParameter) tuples.
        """
        return self._dependencies


class IdentityTransform(AbstractTransformation):

    def __init__(self, *args, **kwargs):
        """The identity transform does no transformation and returns the input given."""
        super(IdentityTransform, self).__init__(*args, **kwargs)

    def get_cl_encode(self):
        return FormatAssignmentConstructor('{parameter_variable}')

    def get_cl_decode(self):
        return FormatAssignmentConstructor('{parameter_variable}')


class ClampTransform(AbstractTransformation):
    """The clamp transformation limits the parameter between its lower and upper bound using the clamp function."""

    def get_cl_encode(self):
        return FormatAssignmentConstructor('clamp((mot_float_type){parameter_variable}, '
                                           '(mot_float_type){lower_bound}, '
                                           '(mot_float_type){upper_bound})')

    def get_cl_decode(self):
        return FormatAssignmentConstructor('clamp((mot_float_type){parameter_variable}, '
                                           '(mot_float_type){lower_bound}, '
                                           '(mot_float_type){upper_bound})')


class CosSqrClampTransform(AbstractTransformation):
    """The clamp transformation limits the parameter between its lower and upper bound using a cos(sqr()) transform."""

    def get_cl_encode(self):
        return FormatAssignmentConstructor(
            'acos(clamp((mot_float_type)sqrt(fabs( ({parameter_variable} - {lower_bound}) / '
            '                                      ({upper_bound} - {lower_bound}) )), '
            '           (mot_float_type)0, (mot_float_type)1))')

    def get_cl_decode(self):
        return FormatAssignmentConstructor('pown(cos({parameter_variable}), 2) * ' +
                                           '({upper_bound} - {lower_bound}) + {lower_bound}')


class SinSqrClampTransform(AbstractTransformation):
    """The clamp transformation limits the parameter between its lower and upper bound using a sin(sqr()) transform."""

    def get_cl_encode(self):
        return FormatAssignmentConstructor(
            'asin(clamp((mot_float_type)sqrt(fabs( ({parameter_variable} - {lower_bound}) / '
            '                                       ({upper_bound} - {lower_bound}) )), '
            '           (mot_float_type)0, (mot_float_type)1))')

    def get_cl_decode(self):
        return FormatAssignmentConstructor('pown(sin({parameter_variable}), 2) * ' +
                                           '({upper_bound} - {lower_bound}) + {lower_bound}')


class SqrClampTransform(AbstractTransformation):
    """The clamp transformation limits the parameter between its lower and upper bound using a sqr() transform."""

    def get_cl_encode(self):
        return FormatAssignmentConstructor('sqrt({parameter_variable})')

    def get_cl_decode(self):
        return FormatAssignmentConstructor('clamp((mot_float_type)({parameter_variable} * {parameter_variable}), '
                                           '      (mot_float_type){lower_bound}, '
                                           '      (mot_float_type){upper_bound})')


class SinSqrClampDependentTransform(AbstractTransformation):
    """The clamp transformation limits the parameter between 0 and the given parameter with the sin(sqr()) transform."""

    def get_cl_encode(self):
        return FormatAssignmentConstructor('asin(sqrt(fabs(({parameter_variable} - {lower_bound}) / '
                                           '               ({dependency_variable_0} - {lower_bound}))))')

    def get_cl_decode(self):
        return FormatAssignmentConstructor('pown(sin({parameter_variable}), 2) * {dependency_variable_0} '
                                           '        + {lower_bound}')


class AbsModXTransform(AbstractTransformation):
    def __init__(self, x, dependencies=()):
        """Create an transformation that returns the absolute modulo x value of the input."""
        super(AbsModXTransform, self).__init__(dependencies)
        self._x = x

    def get_cl_encode(self):
        return FormatAssignmentConstructor('fmod((mot_float_type)fabs({parameter_variable}), '
                                           '(mot_float_type)' + str(self._x) + ')')

    def get_cl_decode(self):
        return FormatAssignmentConstructor('fmod((mot_float_type)fabs({parameter_variable}), '
                                           '(mot_float_type)' + str(self._x) + ')')


class AbsModPiTransform(AbsModXTransform):
    def __init__(self):
        super(AbsModPiTransform, self).__init__('M_PI')


class AssignmentConstructor(object):

    def create_assignment(self, parameter_variable, lower_bound, upper_bound):
        """Create the assignment string.

        Args:
            parameter_variable (str): the name of the parameter variable holding the current value in the kernel
            lower_bound (str): the value or the name of the variable holding the value for the lower bound
            upper_bound (str): the value or the name of the variable holding the value for the upper bound

        Returns:
            str: the transformation assignment
        """


class FormatAssignmentConstructor(AssignmentConstructor):

    def __init__(self, assignment):
        """Assignment constructor that formats the given assignment template.

        This expects that the assignment string has elements like:

        * ``{parameter_variable}``: for the parameter variable
        * ``{lower_bound}``: for the lower bound
        * ``{upper_bound}``: for the upper bound
        * ``{dependency_variable_<n>}: for the dependency variable names

        Args:
            assignment (str): the string containing the assignment template.
        """
        self._assignment = assignment

    def create_assignment(self, parameter_variable, lower_bound, upper_bound, dependency_variables=()):
        assignment = self._assignment.replace('{parameter_variable}', parameter_variable)
        assignment = assignment.replace('{lower_bound}', lower_bound)
        assignment = assignment.replace('{upper_bound}', upper_bound)

        for ind, var_name in enumerate(dependency_variables):
            assignment = assignment.replace('{dependency_variable_' + str(ind) + '}', var_name)

        return assignment



import logging
import pyopencl as cl
import numpy as np
from ...utils import get_float_type_def
from ...cl_routines.base import CLRoutine
from ...load_balance_strategies import Worker


__author__ = 'Robbert Harms'
__date__ = "2014-05-18"
__license__ = "LGPL v3"
__maintainer__ = "Robbert Harms"
__email__ = "robbert.harms@maastrichtuniversity.nl"


class CodecRunner(CLRoutine):

    def __init__(self, **kwargs):
        """This class can run the parameter encoding and decoding transformations.

        These transformations are used to transform the parameters to and from optimization space.
        """
        super(CodecRunner, self).__init__(**kwargs)
        self._logger = logging.getLogger(__name__)

    def decode(self, model, data):
        """Decode the given parameters using the given model.

        This transforms the data from optimization space to model space.

        Args:
            model (mot.model_interfaces.OptimizeModelInterface): The model to use
            data (ndarray): The parameters to transform to model space

        Returns:
            ndarray: The array with the transformed parameters.
        """
        if len(data.shape) > 1:
            from_width = data.shape[1]
        else:
            from_width = 1

        return self._transform_parameters(model.get_parameter_decode_function('decodeParameters'),
                                          'decodeParameters', data,
                                          from_width, model)

    def encode(self, model, data):
        """Encode the given parameters using the given model.

        This transforms the data from model space to optimization space.

        Args:
            model (mot.model_interfaces.OptimizeModelInterface): The model to use
            data (ndarray): The parameters to transform to optimization space

        Returns:
            ndarray: The array with the transformed parameters.
        """
        if len(data.shape) > 1:
            from_width = data.shape[1]
        else:
            from_width = 1

        return self._transform_parameters(model.get_parameter_encode_function('encodeParameters'),
                                          'encodeParameters', data,
                                          from_width, model)

    def _transform_parameters(self, cl_func, cl_func_name, data, nmr_params, model):
        np_dtype = np.float32
        if model.double_precision:
            np_dtype = np.float64

        data = np.require(data, np_dtype, requirements=['C', 'A', 'O', 'W'])

        workers = self._create_workers(lambda cl_environment: _CodecWorker(cl_environment, self.get_compile_flags_list(),
                                                                           cl_func, cl_func_name, data,
                                                                           nmr_params, model))
        self.load_balancer.process(workers, data.shape[0])
        return data


class _CodecWorker(Worker):

    def __init__(self, cl_environment, compile_flags, cl_func, cl_func_name, data, nmr_params, model):
        super(_CodecWorker, self).__init__(cl_environment)
        self._cl_func = cl_func
        self._cl_func_name = cl_func_name
        self._data = data
        self._nmr_params = nmr_params
        self._model = model

        self._param_buf = cl.Buffer(self._cl_run_context.context,
                                    cl.mem_flags.READ_WRITE | cl.mem_flags.USE_HOST_PTR,
                                    hostbuf=self._data)
        self._all_buffers = [self._param_buf]
        for data in self._model.get_data():
            self._all_buffers.append(cl.Buffer(self._cl_run_context.context,
                                               cl.mem_flags.READ_ONLY | cl.mem_flags.COPY_HOST_PTR, hostbuf=data))

        self._kernel = self._build_kernel(compile_flags)

    def __del__(self):
        for buffer in self._all_buffers:
            buffer.release()

    def calculate(self, range_start, range_end):
        nmr_problems = range_end - range_start

        event = self._kernel.transformParameterSpace(self._cl_run_context.queue, (int(nmr_problems), ), None,
                                                     *self._all_buffers, global_offset=(int(range_start),))

        return [self._enqueue_readout(self._param_buf, self._data, range_start, range_end, [event])]

    def _get_kernel_source(self):
        kernel_param_names = ['global mot_float_type* x_global'] + \
                             self._model.get_kernel_param_names(self._cl_environment.device)

        kernel_source = ''
        kernel_source += get_float_type_def(self._model.double_precision)
        kernel_source += str(self._model.get_kernel_data_struct(self._cl_environment.device))
        kernel_source += self._cl_func
        kernel_source += '''
            __kernel void transformParameterSpace(
                ''' + ",\n".join(kernel_param_names) + '''){
                int gid = get_global_id(0);

                ''' + self._model.get_kernel_data_struct_initialization(self._cl_environment.device, 'data') + '''

                mot_float_type x[''' + str(self._nmr_params) + '''];

                for(int i = 0; i < ''' + str(self._nmr_params) + '''; i++){
                    x[i] = x_global[gid * ''' + str(self._nmr_params) + ''' + i];
                }

                ''' + self._cl_func_name + '''((void*)&data, x);

                for(int i = 0; i < ''' + str(self._nmr_params) + '''; i++){
                    x_global[gid * ''' + str(self._nmr_params) + ''' + i] = x[i];
                }
            }
        '''
        return kernel_source



__author__ = 'Robbert Harms'
__date__ = "2014-06-19"
__license__ = "LGPL v3"
__maintainer__ = "Robbert Harms"
__email__ = "robbert.harms@maastrichtuniversity.nl"


class AbstractParameterPrior(object):
    """The priors are used during model sampling.

    They indicate the a priori information one has about a parameter.
    """

    def get_cl_assignment(self):
        """Get the assignment constructor.

        Returns:
            AssignmentConstructor: The assignment construction method for this prior
        """
        return ''


class UniformPrior(AbstractParameterPrior):
    """The uniform prior is always 1."""
    def get_cl_assignment(self):
        return FormatAssignmentConstructor('1')


class UniformWithinBoundsPrior(AbstractParameterPrior):
    """This prior is 1 within the upper and lower bound of the parameter, 0 outside."""
    def get_cl_assignment(self):
        return FormatAssignmentConstructor('(({parameter_variable} < {lower_bound} '
                                           '  || {parameter_variable} > {upper_bound}) ? 0.0 : 1.0)')


class AbsSinPrior(AbstractParameterPrior):
    """The fabs(sin(x)) prior."""
    def get_cl_assignment(self):
        return FormatAssignmentConstructor('fabs(sin({parameter_variable}))')


class AbsSinHalfPrior(AbstractParameterPrior):
    """The fabs(sin(x)/2.0) prior. Taken from FSL"""
    def get_cl_assignment(self):
        return FormatAssignmentConstructor('fabs(sin({parameter_variable})/2.0)')


class AssignmentConstructor(object):

    def create_assignment(self, parameter_variable, lower_bound, upper_bound):
        """Create the assignment string.

        Args:
            parameter_variable (str): the name of the parameter variable holding the current value in the kernel
            lower_bound (str): the value or the name of the variable holding the value for the lower bound
            upper_bound (str): the value or the name of the variable holding the value for the upper bound

        Returns:
            str: the prior assignment
        """


class FormatAssignmentConstructor(AssignmentConstructor):

    def __init__(self, assignment):
        """Assignment constructor that formats the given assignment template.

        This expects that the assignment string has elements like:

        * ``{parameter_variable}``: for the parameter variable
        * ``{lower_bound}``: for the lower bound
        * ``{upper_bound}``: for the upper bound

        Args:
            assignment (str): the string containing the assignment template.
        """
        self._assignment = assignment

    def create_assignment(self, parameter_variable, lower_bound, upper_bound):
        assignment = self._assignment.replace('{parameter_variable}', parameter_variable)
        assignment = assignment.replace('{lower_bound}', lower_bound)
        assignment = assignment.replace('{upper_bound}', upper_bound)
        return assignment



import os
from random import Random

import numpy as np
from pkg_resources import resource_filename

__author__ = 'Robbert Harms'
__date__ = "2016-12-03"
__maintainer__ = "Robbert Harms"
__email__ = "robbert.harms@maastrichtuniversity.nl"


class Random123StartingPoint(object):

    def get_key(self):
        """Gets the key used as starting point for the Random123 generator.

        This should either return 0 or 2 keys depending on the desired precision.

        Returns:
            ndarray: 0 or 2 elements of dtype uint32.
        """

    def get_counter(self):
        """Gets the counter used as starting point for the Random123 generator.

        Returns:
            ndarray: 4 elements of dtype uint32.
        """


class StartingPointFromSeed(Random123StartingPoint):

    def __init__(self, seed, key_length=None):
        """Generates the key and counter from the given seed.

        Args:
            seed (int): the seed used to generate a starting point for the Random123 RNG.
            key_length (int): the length of the key, either 0 or 2 depending on the desired precision.
        """
        self._seed = seed
        self._key_length = key_length or 2
        if self._key_length not in (0, 2):
            raise ValueError('The key length should be either 0 or 2, {} given.'.format(key_length))

    def get_key(self):
        rng = Random(self._seed)
        dtype_info = np.iinfo(np.uint32)

        return np.array(list(rng.randrange(dtype_info.min, dtype_info.max + 1) for _ in range(self._key_length)),
                        dtype=np.uint32)

    def get_counter(self):
        rng = Random(self._seed + 1)
        dtype_info = np.iinfo(np.uint32)

        return np.array(list(rng.randrange(dtype_info.min, dtype_info.max + 1) for _ in range(4)), dtype=np.uint32)


class RandomStartingPoint(StartingPointFromSeed):

    def __init__(self, **kwargs):
        """Generates the key and counter randomly."""
        super(RandomStartingPoint, self).__init__(Random().randint(0, 2**31), **kwargs)


def get_random123_cl_code():
    """Get the source code needed for working with the Rand123 RNG.

    Returns:
        str: the CL code for the Rand123 RNG
    """
    generator = 'threefry'

    src = open(os.path.abspath(resource_filename('mot', 'data/opencl/random123/openclfeatures.h'), ), 'r').read()
    src += open(os.path.abspath(resource_filename('mot', 'data/opencl/random123/array.h'), ), 'r').read()
    src += open(os.path.abspath(
        resource_filename('mot', 'data/opencl/random123/{}.h'.format(generator)), ), 'r').read()
    src += open(os.path.abspath(resource_filename('mot', 'data/opencl/random.h'.format(generator)), ), 'r').read()
    src += (open(os.path.abspath(resource_filename('mot', 'data/opencl/random123/rand123.cl'), ), 'r').read() % {
       'GENERATOR_FUNCTION': (generator + '4x32')
    })

    return src



from mot.model_building.cl_functions.parameters import FreeParameter
from mot.model_building.cl_functions.base import ModelFunction
from mot.model_building.parameter_functions.transformations import CosSqrClampTransform
from mot.cl_data_type import CLDataType

__author__ = 'Robbert Harms'
__date__ = "2014-08-05"
__license__ = "LGPL v3"
__maintainer__ = "Robbert Harms"
__email__ = "robbert.harms@maastrichtuniversity.nl"


class SignalNoiseModel(ModelFunction):

    def __init__(self, name, cl_function_name, parameter_list, dependency_list=()):
        """Signal noise models can add noise to the signal resulting from the model.

        They require the signal resulting from the model and zero or more parameters and they return a new signal
        with noise added.
        """
        super(SignalNoiseModel, self).__init__(name, cl_function_name, parameter_list, dependency_list=dependency_list)

    def get_signal_function(self, fname='signalNoiseModel'):
        """Get the signal function that adds the noise to the signal function.

        Args:
            fname (str, optional): The function name of the function in OpenCL.

        Returns:
            str: A function with signature:

            .. code-block:: c

                mot_float_type fname(const mot_float_type signal, <noise model parameters ...>);

            For example, if the noise model has only one parameter 'sigma' the function should look like:

            .. code-block:: c

                mot_float_type fname(const mot_float_type signal, const mot_float_type sigma);

            The CL function should return a single mot_float_type that represents the signal with the signal noise
                added to it.
        """


class JohnsonSignalNoise(SignalNoiseModel):

    def __init__(self):
        """Johnson noise adds noise to the signal using the formula:

        .. code-block:: c

            sqrt(signal^2 + eta^2)

        """
        super(JohnsonSignalNoise, self).__init__(
            'JohnsonNoise',
            'johnsonNoiseModel',
            (FreeParameter(CLDataType.from_string('mot_float_type'), 'eta', False, 0.1, 0, 1e6,
                           parameter_transform=CosSqrClampTransform()),), ())

    def get_signal_function(self, fname='signalNoiseModel'):
        return '''
            mot_float_type ''' + fname + '''(const mot_float_type signal, const mot_float_type eta){
                return sqrt((signal * signal) + (eta * eta));
            }
        '''



import six

__author__ = 'Robbert Harms'
__date__ = "2015-03-21"
__license__ = "LGPL v3"
__maintainer__ = "Robbert Harms"
__email__ = "robbert.harms@maastrichtuniversity.nl"


class CLDataType(object):

    def __init__(self, raw_data_type, is_pointer_type=False, vector_length=None,
                 address_space_qualifier=None, pre_data_type_type_qualifiers=None, post_data_type_type_qualifier=None):
        """Create a new CL data type container.

        The CL type can either be a CL native type (``half``, ``double``, ``int``, ...) or the
        special ``mot_float_type`` type.

        Args:
            raw_data_type (str): the specific data type without the vector number and asterisks
            is_pointer_type (boolean): If this parameter is a pointer type (appended by a ``*``)
            vector_length (int or None): If None this data type is not a CL vector type.
                If it is an integer it is the vector length of this data type (2, 3, 4, ...)
            address_space_qualifier (str or None): the address space qualifier or None if not used. One of:
                {``__local``, ``local``, ``__global``, ``global``,
                ``__constant``, ``constant``, ``__private``, ``private``} or None.
            pre_data_type_type_qualifiers (list of str or None): the type qualifiers to use before the data type.
                One of {const, restrict, volatile}
            post_data_type_type_qualifier (str or None): the type qualifier to use after the data type.
                Can only be 'const'
        """
        self.raw_data_type = str(raw_data_type)
        self.is_pointer_type = is_pointer_type
        self.vector_length = vector_length

        if self.vector_length:
            self.vector_length = int(self.vector_length)

        self.address_space_qualifier = address_space_qualifier
        self.pre_data_type_type_qualifiers = pre_data_type_type_qualifiers

        if isinstance(self.pre_data_type_type_qualifiers, six.string_types):
            self.pre_data_type_type_qualifiers = [self.pre_data_type_type_qualifiers]

        self.post_data_type_type_qualifier = post_data_type_type_qualifier

    def get_declaration(self):
        declaration = ''
        if self.address_space_qualifier:
            declaration += str(self.address_space_qualifier) + ' '
        if self.pre_data_type_type_qualifiers:
            declaration += str(' '.join(self.pre_data_type_type_qualifiers)) + ' '
        declaration += str(self.cl_type)
        if self.post_data_type_type_qualifier:
            declaration += ' ' + str(self.post_data_type_type_qualifier)
        return declaration

    @classmethod
    def from_string(cls, parameter_declaration):
        """Parse the parameter declaration into a CLDataType

        Args:
            parameter_declaration (str): the CL parameter declaration. Example: ``global const float4*`` const

        Returns:
            mot.cl_data_type.CLDataType: the CL data type for this parameter declaration
        """
        from mot.parsers.cl.CLDataTypeParser import parse
        return parse(parameter_declaration)

    @property
    def cl_type(self):
        """Get the type of this parameter in CL language

        This only returns the parameter type (like ``double`` or ``int*`` or ``float4*`` ...). It does not include other
        qualifiers.

        Returns:
            str: The name of this data type
        """
        s = self.raw_data_type

        if self.vector_length is not None:
            s += str(self.vector_length)

        if self.is_pointer_type:
            s += '*'

        return str(s)

    @property
    def is_vector_type(self):
        """Check if this data type is a vector type

        Returns:
            boolean: True if it is a vector type, false otherwise
        """
        return self.vector_length is not None

    def set_vector_length(self, vector_length):
        """Set the vector length of this data type.

        Args:
            vector_length (int or None): If None this data type is not a CL vector type.
                If it is an integer it is the vector length of this data type (2, 3, 4, ...)
        """
        self.vector_length = vector_length

    def set_is_pointer_type(self, is_pointer_type):
        """Set if this parameter should be treated a being a pointer type

        Args:
            is_pointer_type (boolean): If this parameter is a pointer type (appened by a ``*``)
        """
        self.is_pointer_type = is_pointer_type

    def set_raw_data_type(self, raw_data_type):
        """Set the raw data type

        Args:
            raw_data_type (str): the specific data type without the vector number and asterisks
        """
        self.raw_data_type = raw_data_type
        return self

    def set_address_space_qualifier(self, address_space_qualifier):
        """Set the address space qualifier.

        Args:
            address_space_qualifier (str): the new address space qualifier

        Returns:
            self: for chaining
        """
        self.address_space_qualifier = address_space_qualifier
        return self

    def set_pre_data_type_type_qualifiers(self, pre_data_type_type_qualifiers):
        """Set the pre data type type qualifier.

        Args:
            pre_data_type_type_qualifiers (list of str): the pre data type type qualifiers

        Returns:
            self: for chaining
        """
        self.pre_data_type_type_qualifiers = pre_data_type_type_qualifiers

        if isinstance(self.pre_data_type_type_qualifiers, six.string_types):
            self.pre_data_type_type_qualifiers = [self.pre_data_type_type_qualifiers]

        return self

    def set_post_data_type_type_qualifier(self, post_data_type_type_qualifier):
        """Set the post data type type qualifier.

        Args:
            post_data_type_type_qualifier (str): the post data type type qualifier

        Returns:
            self: for chaining
        """
        self.post_data_type_type_qualifier = post_data_type_type_qualifier
        return self

    def __str__(self):
        return self.get_declaration()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
test_mot
----------------------------------

Tests for `mot` module.
"""

import unittest

import numpy as np

import mot
from mot import configuration
from mot.cl_routines.mapping.residual_calculator import ResidualCalculator
from mot.cl_routines.optimizing.nmsimplex import NMSimplex
from mot.cl_routines.optimizing.levenberg_marquardt import LevenbergMarquardt
from mot.cl_routines.optimizing.powell import Powell
from mot.cl_routines.filters.gaussian import GaussianFilter
from mot.cl_routines.filters.mean import MeanFilter
from mot.cl_routines.filters.median import MedianFilter
from mot.model_building.models_examples import Rosenbrock, MatlabLSQNonlinExample


class CLRoutineTestCase(unittest.TestCase):

    def __init__(self, *args, **kwargs):
        super(CLRoutineTestCase, self).__init__(*args, **kwargs)
        mot.configuration._config['compile_flags']['general'].update({
            '-cl-single-precision-constant': False
        })


class TestRosenbrock(CLRoutineTestCase):

    def setUp(self):
        self.model = Rosenbrock(5)
        self.optimizers = (NMSimplex(), Powell(patience=10))

    def test_model(self):
        for optimizer in self.optimizers:
            v = optimizer.minimize(self.model)
            for p in self.model.get_optimized_param_names():
                self.assertAlmostEqual(float(v[p]), 1.0, places=3)


class TestLSQNonLinExample(CLRoutineTestCase):

    def setUp(self):
        self.model = MatlabLSQNonlinExample()
        self.optimizers = (LevenbergMarquardt(),)
        self.residual_calc = ResidualCalculator()

    def test_model(self):
        for optimizer in self.optimizers:
            v = optimizer.minimize(self.model)
            res = self.residual_calc.calculate(self.model, v)
            s = 0
            for i in range(res.shape[1]):
                s += res[0, i]**2
            self.assertAlmostEqual(s, 124.3622, places=4)


class TestFilters(CLRoutineTestCase):

    def setUp(self):
        self.d1 = np.array([1, 2, 4, 2, 1], dtype=np.float64)
        self.d2 = np.eye(4)

    def test_median(self):
        filter = MedianFilter(2)
        s1 = filter.filter(self.d1)
        np.testing.assert_almost_equal(s1, np.array([2, 2, 2, 2, 2]))

        s2 = filter.filter(self.d2)
        np.testing.assert_almost_equal(s2, np.zeros((4, 4)))

    def test_mean(self):
        filter = MeanFilter(2)
        s1 = filter.filter(self.d1)
        np.testing.assert_almost_equal(s1, np.array([2 + 1/3.0, 2.25, 2, 2.25, 2 + 1/3.0]))

        s2 = filter.filter(self.d2)
        expected = np.ones((4, 4)) * 0.25
        expected[0, 0] = 1/3.0
        expected[0, 3] = 2/9.0
        expected[3, 0] = 2/9.0
        expected[3, 3] = 1/3.0
        np.testing.assert_almost_equal(s2, expected)

    def test_gaussian(self):
        filter = GaussianFilter(2, sigma=1.0)
        s1 = filter.filter(self.d1, mask=np.array([1, 1, 1, 1, 0]))
        s2 = filter.filter(self.d2)

        np.testing.assert_almost_equal(s1, [1.1089774, 2.135224, 2.6417738, 1.8910226, 0])

        expected = np.array([[0.22470613, 0.20994687, 0.10351076, 0.02661242],
                             [0.20994687, 0.28434043, 0.22325308, 0.10351076],
                             [0.10351076, 0.22325308, 0.28434043, 0.20994687],
                             [0.02661242, 0.10351076, 0.20994687, 0.22470613]])
        np.testing.assert_almost_equal(s2, expected)


if __name__ == '__main__':
    unittest.main()



import os
from pkg_resources import resource_filename
from .base import AbstractParallelOptimizer, AbstractParallelOptimizerWorker

__author__ = 'Robbert Harms'
__date__ = "2014-02-05"
__license__ = "LGPL v3"
__maintainer__ = "Robbert Harms"
__email__ = "robbert.harms@maastrichtuniversity.nl"


class Powell(AbstractParallelOptimizer):

    default_patience = 2

    def __init__(self, bracket_gold=None, glimit=None, reset_method=None, patience=None, optimizer_settings=None,
                 **kwargs):
        """Use the Powell method to calculate the optimum.

        Args:
            patience (int):
                Used to set the maximum number of iterations to patience*(number_of_parameters+1)
            bracket_gold (double): the default ratio by which successive intervals are magnified in Bracketing
            glimit (double): the maximum magnification allowed for a parabolic-fit step in Bracketing
            reset_method (str): one of 'EXTRAPOLATED_POINT' or 'RESET_TO_IDENTITY' lower case or upper case.
        """
        patience = patience or self.default_patience

        optimizer_settings = optimizer_settings or {}

        keyword_values = {}
        keyword_values['bracket_gold'] = bracket_gold
        keyword_values['glimit'] = glimit
        keyword_values['reset_method'] = reset_method

        option_defaults = {'bracket_gold': 1.618034, 'glimit': 100.0, 'reset_method': 'EXTRAPOLATED_POINT'}

        def get_value(option_name):
            value = keyword_values.get(option_name)
            if value is None:
                value = optimizer_settings.get(option_name)
            if value is None:
                value = option_defaults[option_name]
            return value

        for option in option_defaults:
            optimizer_settings.update({option: get_value(option)})

        super(Powell, self).__init__(patience=patience, optimizer_settings=optimizer_settings, **kwargs)

    def _get_worker_generator(self, *args):
        return lambda cl_environment: PowellWorker(cl_environment, *args)


class PowellWorker(AbstractParallelOptimizerWorker):

    def _get_optimization_function(self):
        params = {'NMR_PARAMS': self._nmr_params, 'PATIENCE': self._parent_optimizer.patience}

        for option, value in self._optimizer_settings.items():
            params.update({option.upper(): value})
        params['RESET_METHOD'] = 'POWELL_RESET_METHOD_' + params['RESET_METHOD'].upper()

        body = open(os.path.abspath(resource_filename('mot', 'data/opencl/powell.pcl')), 'r').read()
        if params:
            body = body % params
        return body

    def _get_optimizer_call_name(self):
        return 'powell'



import numpy as np
from mot.cl_routines.optimizing.base import AbstractOptimizer

__author__ = 'Robbert Harms'
__date__ = "2016-11-22"
__maintainer__ = "Robbert Harms"
__email__ = "robbert.harms@maastrichtuniversity.nl"


class StartingPointGenerator(object):

    def next(self, model, previous_results):
        """Returns a next starting point given the model and the previous results.

        This method can return None which means that no next starting point is available.

        Args:
            model: the model for which we are generating points
            previous_results (dict): the dictionary with per parameter the results

        Returns:
            dict: per parameter a ndarray with the new starting points per problem instance.
        """
        return NotImplementedError


class PointsFromGrid(StartingPointGenerator):

    def __init__(self, grid):
        """Uses the given points as a starting point for all problem instances per run.

        This accepts a grid with per row a starting point that we will use for all the problem instances for
        that optimization run. The number of rows determines the number of iterations.

        Args:
            grid (ndarray): the grid with one starting point per iteration.
        """
        self.grid = grid
        self._iteration_counter = 0

    def next(self, model, previous_results):
        if self._iteration_counter == self.grid.shape[0]:
            return None

        init_dict = {}

        for param_ind, param_name in enumerate(model.get_optimized_param_names()):
            init_dict[param_name] = np.ones(model.get_nmr_problems()) * self.grid[self._iteration_counter, param_ind]

        self._iteration_counter += 1
        return init_dict


class UsePrevious(StartingPointGenerator):

    def __init__(self, number_of_runs=3):
        """A strategy that uses the previous results without alterations for the next optimization run.

        Args:
            number_of_runs (int): the number of times we optimize using this strategy
        """
        self.number_of_runs = number_of_runs
        self._iteration_counter = 0

    def next(self, model, previous_results):
        if self._iteration_counter == self.number_of_runs:
            return None
        self._iteration_counter += 1
        return previous_results


class RandomStartingPoint(StartingPointGenerator):

    def __init__(self, number_of_runs=3):
        """A strategy that generates uniformly random starting points for every parameter.

        Per run this generates for each parameter a uniformly distributed random number between the lower
        and upper bounds and uses that single random value for all problem instances.

        Hence, this does not generate a unique random point per problem instance, but uses a single random
        point for all problem instances per iteration.

        Args:
            number_of_runs (int): the number of times we optimize using this strategy
        """
        self.number_of_runs = number_of_runs
        self._iteration_counter = 0

    def next(self, model, previous_results):
        if self._iteration_counter == self.number_of_runs:
            return None
        self._iteration_counter += 1

        param_names = model.get_optimized_param_names()
        lower_bounds = model.get_lower_bounds()
        upper_bounds = model.get_upper_bounds()

        init_dict = {}

        for param_ind, param_name in enumerate(param_names):
            init_dict[param_name] = np.ones(model.get_nmr_problems()) * \
                                    np.random.uniform(np.min(lower_bounds[param_ind]),
                                                      np.max(upper_bounds[param_ind]))

        return init_dict


class GaussianPerturbation(StartingPointGenerator):

    def __init__(self, number_of_runs=3):
        """A strategy that perturbates the previous results using a Normal distribution

        Per run this generates for each parameter and for each problem instance a new starting position using
        the previous parameter as a mean and for standard deviation it uses the standard deviation over the problem
        instances.

        Hence, this generates a unique random point per problem instance, in contrast to some of the other strategies.

        Args:
            number_of_runs (int): the number of times we optimize using this strategy
        """
        self.number_of_runs = number_of_runs
        self._iteration_counter = 0

    def next(self, model, previous_results):
        if self._iteration_counter == self.number_of_runs:
            return None
        self._iteration_counter += 1

        init_dict = {}

        for param_ind, param_name in enumerate(model.get_optimized_param_names()):
            std = np.std(previous_results[param_name])

            points = np.zeros(model.get_nmr_problems())
            for problem_ind in range(model.get_nmr_problems()):
                points[problem_ind] = np.random.normal(previous_results[param_name][problem_ind], std)

            init_dict[param_name] = points

        return init_dict


class RandomRestart(AbstractOptimizer):

    def __init__(self, optimizer, starting_point_generator, **kwargs):
        """A meta optimization routine that allows multiple random restarts.

        This meta optimizer runs the given optimizer multiple times using different starting positions for each run.
        The starting positions are obtained using a starting point generator which returns new starting points
        given the initial starting points.

        Please note that the initial starting point is always optimized first as a baseline reference.

        The returned results contain per problem instance the parameter that resulted in the lowest function value.

        Args:
            optimizer (AbstractOptimizer): the optimization routines to run one after another.
            starting_point_generator (StartingPointGenerator): the randomizer instance we use to randomize the starting point
        """
        super(RandomRestart, self).__init__(**kwargs)
        self._optimizer = optimizer
        self._starting_point_generator = starting_point_generator

    def minimize(self, model, init_params=None, full_output=False):
        results = self._optimizer.minimize(model, init_params, full_output=True)

        starting_points = self._starting_point_generator.next(model, results[0])
        while starting_points:
            new_results = self._optimizer.minimize(model, starting_points, full_output=True)

            results = self._get_best_results(results, new_results)

            starting_points = self._starting_point_generator.next(model, results[0])

        return results

    def _get_best_results(self, previous_results, new_results):
        result_choice = np.argmin([previous_results[1]['Errors.l2'], new_results[1]['Errors.l2']], axis=0)

        output = []
        for results_ind in range(len(previous_results)):
            output_dict = {}

            for map_name in previous_results[results_ind]:
                output_dict[map_name] = np.array([previous_results[results_ind][map_name],
                          new_results[results_ind][map_name]])[(result_choice, range(result_choice.shape[0]))]

            output.append(output_dict)

        return output



from collections import Mapping

import pyopencl as cl
import numpy as np
from ...utils import get_float_type_def
from ...cl_routines.base import CLRoutine
from ...load_balance_strategies import Worker


__author__ = 'Robbert Harms'
__date__ = "2014-02-05"
__license__ = "LGPL v3"
__maintainer__ = "Robbert Harms"
__email__ = "robbert.harms@maastrichtuniversity.nl"


class ObjectiveCalculator(CLRoutine):

    def calculate(self, model, parameters):
        """Calculate and return the objective function of the given model for the given parameters.

        This evaluates the model and compares it to the problem data to get objective values. This returns
        the objective value per problem instance, for an objective function value per observation per problem use
        the :class:`~.objective_list_calculator.ObjectiveListCalculator`.

        Args:
            model (AbstractModel): The model to calculate the objective function of.
            parameters (dict or ndarray): The parameters to use in the evaluation of the model
                If a dict is given we assume it is with values for a set of parameters
                If an ndarray is given we assume that we have data for all parameters.

        Returns:
            Return per voxel the objective function value
        """
        parameters = self._initialize_parameters(parameters, model)
        objective_values = self._initialize_result_array(model)

        workers = self._create_workers(
            lambda cl_environment: _ObjectiveCalculatorWorker(cl_environment, self.get_compile_flags_list(),
                                                              model, parameters, objective_values))
        self.load_balancer.process(workers, model.get_nmr_problems())

        return objective_values

    def _initialize_parameters(self, parameters, model):
        np_dtype = np.float32
        if model.double_precision:
            np_dtype = np.float64

        if isinstance(parameters, Mapping):
            return np.require(model.get_initial_parameters(parameters), np_dtype, requirements=['C', 'A', 'O'])

        return np.require(parameters, np_dtype, requirements=['C', 'A', 'O'])

    def _initialize_result_array(self, model):
        nmr_problems = model.get_nmr_problems()
        np_dtype = np.float32
        if model.double_precision:
            np_dtype = np.float64
        return np.zeros((nmr_problems,), dtype=np_dtype, order='C')


class _ObjectiveCalculatorWorker(Worker):

    def __init__(self, cl_environment, compile_flags, model, parameters, objective_values):
        super(_ObjectiveCalculatorWorker, self).__init__(cl_environment)

        self._model = model
        self._double_precision = model.double_precision
        self._objective_values = objective_values
        self._parameters = parameters

        self._all_buffers, self._objective_values_buffer = self._create_buffers()
        self._kernel = self._build_kernel(compile_flags)

    def __del__(self):
        for buffer in self._all_buffers:
            buffer.release()

    def calculate(self, range_start, range_end):
        nmr_problems = range_end - range_start
        event = self._kernel.run_kernel(self._cl_run_context.queue, (int(nmr_problems), ), None, *self._all_buffers,
                                        global_offset=(int(range_start),))
        return [self._enqueue_readout(self._objective_values_buffer, self._objective_values,
                                      range_start, range_end, [event])]

    def _create_buffers(self):
        objective_value_buffer = cl.Buffer(self._cl_run_context.context,
                                           cl.mem_flags.WRITE_ONLY | cl.mem_flags.USE_HOST_PTR,
                                           hostbuf=self._objective_values)

        params_buffer = cl.Buffer(self._cl_run_context.context,
                                  cl.mem_flags.READ_ONLY | cl.mem_flags.USE_HOST_PTR,
                                  hostbuf=self._parameters)

        all_buffers = [params_buffer, objective_value_buffer]

        for data in self._model.get_data():
            all_buffers.append(cl.Buffer(self._cl_run_context.context,
                                         cl.mem_flags.READ_ONLY | cl.mem_flags.COPY_HOST_PTR, hostbuf=data))

        return all_buffers, objective_value_buffer

    def _get_kernel_source(self):
        nmr_params = self._parameters.shape[1]

        kernel_param_names = ['global mot_float_type* params', 'global mot_float_type* objective_values']
        kernel_param_names.extend(self._model.get_kernel_param_names(self._cl_environment.device))
        kernel_source = ''
        kernel_source += get_float_type_def(self._double_precision)
        kernel_source += self._model.get_kernel_data_struct(self._cl_environment.device)
        kernel_source += self._model.get_objective_function('calculateObjective')
        kernel_source += '''
            __kernel void run_kernel(
                ''' + ",\n".join(kernel_param_names) + '''
                ){
                    int gid = get_global_id(0);
                    ''' + self._model.get_kernel_data_struct_initialization(self._cl_environment.device, 'data') + '''

                    mot_float_type x[''' + str(nmr_params) + '''];
                    for(int i = 0; i < ''' + str(nmr_params) + '''; i++){
                        x[i] = params[gid * ''' + str(nmr_params) + ''' + i];
                    }

                    objective_values[gid] = calculateObjective((void*)&data, x);
            }
        '''
        return kernel_source



__author__ = 'Robbert Harms'
__date__ = "2014-06-19"
__license__ = "LGPL v3"
__maintainer__ = "Robbert Harms"
__email__ = "robbert.harms@maastrichtuniversity.nl"


class AbstractParameterDependency(object):

    @property
    def pre_transform_code(self):
        """Some code that may be prefixed to this parameter dependency.

        Here one can put more elaborate CL code. Please make sure that additional variables are unique.

        Returns:
            str: The pre transformation code. This is prepended to the dependency function.
        """
        return ''

    @property
    def assignment_code(self):
        """Get the assignment code (including a ;).

        Returns:
            str: The assignment code.
        """
        return ''

    @property
    def fixed(self):
        """Check if this dependency fixes the parameter to the assigned value, or if it is still estimable.

        Returns:
            boolean: True if this parameter is now fixed to the dependency, or false if it is not.
        """
        return True

    @property
    def has_side_effects(self):
        """Check if the pre_transform_code from this parameter dependency has side effects to other parameters.

        Returns:
            boolean: True if this parameter has side effects (in the pre_transform_code) or does not have side-effects.
        """
        return False


class SimpleAssignment(AbstractParameterDependency):

    def __init__(self, assignment_code, fixed=True, has_side_effects=False):
        """Adds a simple parameter dependency rule for the given parameter.

        This is for one parameter, a simple one-line transformation dependency.

        Args:
            assignment_code (str): the assignment code (in CL) for this parameter
            fixed (boolean): if this parameters fixes to the assigned value or not
            has_side_effects (boolean): if this parameter changes one of the parameters before it goes into the model
                functions. Note that these side effects must be idempotent.
        """
        self._assignment = assignment_code
        self._fixed = fixed
        self._has_side_effects = has_side_effects

    @property
    def assignment_code(self):
        return self._assignment

    @property
    def fixed(self):
        return self._fixed

    @property
    def has_side_effects(self):
        return self._has_side_effects


class WeightSumToOneRule(AbstractParameterDependency):

    def __init__(self, parameter_names):
        """Adds the unity sum (parameter) dependency for the weights indicated by the given parameters.

        Parameters are given by (by <model>.<param> name). This makes sure that the given parameters sum to one before
        they are used in the model functions. Note that if you have 3 weights, you make a dependency for one of them
        with as argument the name of the other two.

        Args:
            parameter_names (list): A list with the the names of the parameters used in the dependency.
        """
        self._pre_transform_code = ''

        if len(parameter_names) < 2:
            self._assignment = '1 - ' + parameter_names[0]
            self._has_side_effects = False
        else:
            divisors = ''
            for p in parameter_names:
                divisors += p + ' /= _weight_sum;' + "\n" + "\t" * 5
            self._pre_transform_code += '''
                mot_float_type _weight_sum = ''' + ' + '.join(parameter_names) + ''';
                if(_weight_sum < 1.0){
                    _weight_sum = 1 - _weight_sum;
                }
                else{
                    ''' + divisors + '''
                    _weight_sum = 0;
                }
            '''
            self._assignment = '_weight_sum'
            self._has_side_effects = True

    @property
    def pre_transform_code(self):
        return self._pre_transform_code

    @property
    def assignment_code(self):
        return self._assignment

    @property
    def fixed(self):
        return True

    @property
    def has_side_effects(self):
        return self._has_side_effects



import numbers
import numpy as np
from pyopencl import array as cl_array

__author__ = 'Robbert Harms'
__date__ = "2016-12-06"
__maintainer__ = "Robbert Harms"
__email__ = "robbert.harms@maastrichtuniversity.nl"


class DataAdapter(object):
    """Create a data adapter for the given data and type.

    This data adapter is the bridge between the raw data and the data used in the kernels. If in the future we
    want to add computation types like CUDA or compiling to C this data adapter can be used as the bridge.
    """

    def get_data_type(self):
        """Get the data type for the data in this adapter.

        Returns:
            mot.cl_data_type.CLDataType: the datatype
        """

    def get_opencl_data(self):
        """Adapt and return the data for use in OpenCL kernels.

        Returns:
            np.ndarray: the data to be used in compute kernels.
        """

    def get_opencl_numpy_type(self):
        """Get the numpy type for the data in this class for OpenCL use.

        Returns:
            np.dtype: the numpy type for the data
        """


class SimpleDataAdapter(DataAdapter):

    def __init__(self, data, data_type, mot_float_type):
        """Create a data adapter for the given data and type.

        This data adapter is the bridge between the raw data and the data used in the kernels. If in the future we
        want to add computation types like CUDA or plain C, this adapter knows how to format the data for those targets.

        Args:
            value (ndarray): The value to adapt to different run environments
            data_type (mot.cl_data_type.CLDataType): the data type we need to convert it to
            mot_float_type (mot.cl_data_type.CLDataType): the data type of the mot_float_type
        """
        self._data = data
        self._data_type = data_type
        self._mot_float_type = mot_float_type

    def get_opencl_data(self):
        if self._data_type.is_vector_type:
            return self._array_to_cl_vector()
        else:
            return self._get_cl_array()

    def get_opencl_numpy_type(self):
        if self._data_type.is_vector_type:
            return self._get_opencl_vector_data_type()
        else:
            return self._get_cl_numpy_type(self._data_type)

    def get_data_type(self):
        return self._data_type

    def _get_cl_array(self):
        """Convert the data to a numpy array of the current data type.

        Returns:
            np.ndarray: the converted data as a numpy type
        """
        numpy_type = self._get_cl_numpy_type(self._data_type)
        if isinstance(self._data, numbers.Number):
            return numpy_type(self._data)
        return np.require(self._data, numpy_type, ['C', 'A', 'O'])

    def _get_cl_numpy_type(self, data_type):
        """Get the numpy data type for non-vector types in CL.

        This function is not part of the CLDataType class since the numpy datatype may differ depending
        on the final use case.

        Args:
            data_type (mot.cl_data_type.CLDataType): the data type to convert to an numpy type
        """
        raw_type = data_type.raw_data_type

        if raw_type == 'float':
            return np.float32
        if raw_type == 'int':
            return np.int32
        if raw_type == 'double':
            return np.float64
        if raw_type == 'mot_float_type':
            return self._get_cl_numpy_type(self._mot_float_type)

    def _array_to_cl_vector(self):
        """Create a CL vector type of the given array.

        Returns:
            ndarray: An array of the same length as the given array, but with only one column per row, the opencl vector.
        """
        s = self._data.shape
        if len(s) > 1:
            width = s[1]
        else:
            width = 1

        dtype = self._get_opencl_vector_data_type()

        ve = np.zeros((s[0], 1), dtype=dtype, order='C')
        for i in range(s[0]):
            for j in range(width):
                ve[i, 0][j] = self._data[i, j]

        return ve

    def _get_opencl_vector_data_type(self):
        """Get the data type for a vector of the given length and given type.

        Returns:
            The vector type given the given vector length and data type
        """
        s = self._data.shape
        if len(s) > 1:
            vector_length = s[1]
        else:
            vector_length = 1

        if self._data_type.raw_data_type == 'double':
            data_type = 'double'
        elif self._data_type.raw_data_type == 'mot_float_type':
            data_type = self._mot_float_type.raw_data_type
        else:
            data_type = 'float'

        if vector_length not in (2, 3, 4, 8, 16):
            raise ValueError('The given vector length is not one of (2, 3, 4, 8, 16)')
        if data_type not in ('char', 'uchar', 'short', 'ushort', 'int', 'uint', 'long', 'ulong', 'float', 'double', 'half'):
            raise ValueError('The given data type ({}) is not supported.'.format(data_type))

        return getattr(cl_array.vec, data_type + str(vector_length))



import logging
from ...cl_routines.base import CLRoutine

__author__ = 'Robbert Harms'
__date__ = "2014-05-18"
__license__ = "LGPL v3"
__maintainer__ = "Robbert Harms"
__email__ = "robbert.harms@maastrichtuniversity.nl"


class AbstractSampler(CLRoutine):

    def __init__(self, cl_environments=None, load_balancer=None, **kwargs):
        super(AbstractSampler, self).__init__(cl_environments=cl_environments, load_balancer=load_balancer, **kwargs)
        self._logger = logging.getLogger(__name__)

    def sample(self, model, init_params=None, full_output=False):
        """Minimize the given model with the given codec using the given environments.

        Args:
            model (SampleModelInterface): the model to minimize
            init_params (dict): a dictionary containing the results of a previous run, provides the starting point
            full_output (boolean): If true, also return other output parameters:
                (samples, other_output_maps). Else only return the samples.

        Returns:
            if full output is False return only the samples in a dictionary. If full output is true it returns a list
            with as first elements the samples dict and as second an output map dict.
        """
